"""Module containing GettableParameterBase class"""
from __future__ import annotations

from abc import abstractmethod
import copy
from typing import Type, TYPE_CHECKING
import logging
import numpy as np
import warnings

from qcodes.validators import Arrays
from qcodes.parameters import ParameterWithSetpoints
from qm import qua

if TYPE_CHECKING:
    from arbok_driver.measurement import Measurement
    from arbok_driver.read_sequence import ReadSequence
    from numpy import ndarray
    from qm.qua._dsl.stream_processing.stream_processing import ResultStreamSource
    from qcodes.dataset.data_set_protocol import ValuesType
    from qcodes.parameters.parameter_base import ParameterBase

class GettableParameterBase(ParameterWithSetpoints):
    """
    This is a valid Gettable not because of inheritance, but because it has the
    expected attributes and methods.
    TODO: zip all streams together, stream at once and unpack while opx resume
    TODO: interleaved measurements

    Attributes:
        unit (str): Unit of the parameter
        label (label): Label for parameter (printed on axis)
        readout (Readout): `Readout` instance that created this parameter
        read_sequence (ReadSequence): `ReadSequence` managing the sequence of
            QUA program

    """
    qua_stream: ResultStreamSource
    def __init__(
            self,
            name: str,
            read_sequence: ReadSequence,
            var_type: Type[int | bool | qua.fixed],
            **kwargs
            ) -> None:
        """
        Constructor class for ReadSequence class
        Args:
            name (dict): name of the GettableParameter
            readout (Readout): Readout class summarizing data streams and variables
        """
        super().__init__(name, vals=Arrays(shape=(1,)), **kwargs)
        self.vals: Arrays
        self.var_type: Type[int | bool | qua.fixed] = var_type
        self.read_sequence: ReadSequence = read_sequence
        self.read_sequence.add_gettable(self)
        self.measurement: Measurement = read_sequence.measurement
        measurement_full_name = f"{self.measurement.driver.short_name}_" \
            f"{self.measurement.short_name}_"
        self._register_name = self.full_name.split(measurement_full_name)[-1]
        
        self.reset_measuerement_attributes()
        self.buffer: None = None
        self.is_mock: bool = self.measurement.is_mock

    @abstractmethod
    def qua_declare_variables(self) -> None:
        """Declares the qua variables and streams for this gettable"""
        pass

    @abstractmethod
    def qua_save_variables(self) -> None:
        """Saves acquired results to qua stream"""
        pass

    def qua_save_streams(self) -> None:
        """Saves acquired results to qua stream"""
        buffer = self.qua_stream.buffer(*self.vals.shape)
        buffer.save(self.full_name)

    def set_raw(self, *args, **kwargs) -> None:
        """Empty abstract `set_raw` method. Parameter not meant to be set"""
        raise NotImplementedError("GettableParameters are not meant to be set")

    def reset_measuerement_attributes(self):
        """Resets all job specific attributes"""
        self.buffer = None
        self.qua_stream = None

    def configure_from_measurement(self, setpoints: tuple[ParameterBase, ...]) -> None:
        """
        Configures the gettable parameter from the measurement object.
        This method sets the sweep dimensions, batch size, and snaked shape
        based on the sweeps defined in the measurement.

        Args:
            setpoints (Tuple[ParameterBase, ...]): The setpoint parameters for this gettable
        """
        self.setpoints = setpoints
        self.vals = Arrays(
            shape = tuple(len(param.get()) for param in self.setpoints))
        self.is_mock = self.measurement.is_mock
    
    def set_qm_buffer(self, qm_job: 'RunningQmJob') -> None:
        """
        Fetches the QM buffer result from the QM driver and assigns it to the
        `buffer` attribute.
        
        Args:
            qm_driver (RunningQmJob): The QM driver instance to fetch the buffer from.
        """
        buffer_name = self.full_name
        if buffer_name not in qm_job.result_handles.keys():
            raise ValueError(
                f"Buffer {buffer_name} not found in QM job result handles. "
                f"Use one of the available ones:\n {qm_job.result_handles.keys()}"
            )
        self.buffer = getattr(qm_job.result_handles, buffer_name)

    def get_raw(self) -> np.ndarray:
        """ 
        Get method to retrieve a single batch of data from a running measurement
        On its first call, the gettables attributes get configured regarding
        the given measurement and the underlying hardware.
        """
        warnings.warn(
            "Directly calling get_raw on gettable (%s)"
            "Make sure the qua program has run and the buffer to fetch is full"
            )
        return self.fetch_results()

    def fetch_results(self) -> np.ndarray:
        """
        Get method to retrieve the results from the OPX buffer.

        Returns:
            np.ndarray: Reshaped data array from the OPX buffer.
        """
        if not self.is_mock:
            data = self._fetch_opx_buffer()
        else:
            logging.debug(
                "GettableParameter %s is in mock mode, returning random array",
                self.name
            )
            data = self._generate_synthetic_data()
        if data is None:
            raise RuntimeError(
                f"GettableParameter {self.name} has no data in the buffer. "
                "Make sure the QUA program has run and the buffer is full."
            )
        if any([x.snake_scan for x in self.measurement.sweeps]):
            raise NotImplementedError(
                "Snaked data reshaping is not yet implemented."
            )
        return data

    def _fetch_opx_buffer(self):
        """
        Fetches the OPX buffer into the `buffer_val` and increments the internal
        counter `count`
        """
        buffer_val = np.array(self.buffer.fetch_all(), dtype = float)
        return buffer_val

    def unpack_self(self, value: ValuesType) -> list[tuple[ParameterBase, ValuesType]]:
        """
        Unpacks the ParameterWithSetpoints, its setpoints and any inferred
        parameters controlled by the setpoints.
        Args:
            value(ValuesType): The data acquired from this parameter.
        Returns:
            A list of tuples of parameters and values to be added as results
            to the dataset.
        """
        unpacked_results: list[tuple[ParameterBase, ValuesType]] = []
        setpoint_params = list(self.setpoints)
        setpoint_data = [param.get() for param in setpoint_params]
        output_grids = list(np.meshgrid(*setpoint_data, indexing="ij"))
        for i, param in enumerate(setpoint_params[:]):
            for inferred_param in param.has_control_of:
                copy_setpoint_data = setpoint_data[:]
                copy_setpoint_data[i] = inferred_param.get()
                setpoint_params.append(inferred_param)
                output_grids.append(
                    np.meshgrid(*copy_setpoint_data, indexing="ij")[i]
                )
        unpacked_results = list(zip(setpoint_params, output_grids))
        unpacked_results.extend(
            [(self, value)]
        )  # Must come last to preserve original ordering
        return unpacked_results
    
    def get_mock_result(self):
        """
        Generates linearly increasing results filling the shape of the batch
        """
        mock_result = np.linspace(0, 1, num=np.prod(list(self.vals.shape)))
        mock_result = mock_result.reshape(self.vals.shape)
        return mock_result
    
    def get_full_dims_and_coords(
            self, ext_dims: list[str], ext_coords: dict[str, ndarray]
        ) -> tuple[list[str], dict[str, ndarray]]:
        """
        Returns the full dims and coordinates for xarray DataArray generation

        Args:
            ext_dims (list): list of external dimesions
            external_coords (dict): 
        """
        dims = copy.copy(ext_dims)
        coords = copy.copy(ext_coords)
        for param in self.setpoints:
            dims.append(param.register_name)
            name = param.register_name
            coords[name] = (name, param.get())
            for inferred_param in param.depends_on:
                coords[inferred_param] = (name, inferred_param.get())
        return dims, coords



from typing import Optional
import logging
import math

from qm import qua

from arbok_driver.parameter_types import ParameterMap, Voltage, Time

def ramp(
        elements: list,
        target: ParameterMap,
        operation: str,
        duration: Time | None = None,
        reference: Optional[ParameterMap | None] = None,
        do_align: bool = True,
        no_play_tolerance: float = 1e-6
    ):
    """ 
    Helper function that `play`s a qua operation on the respective elements 
    specified in the sequence config.

    Args:
        elements (list): elements on which pulse is applied
        target (ParameterMap): voltage point to move to
        operation (str): Operation to be played -> find in OPX config
        reference (Optional[ParameterMap | None]): voltage point to come from
        duration (str | qcodes.Parameter): duration of the operation 
        do_align (optional, bool): whether to align elements before and after
            the ramp
        no_play_tolerance (optional, float): tolerance for not playing a pulse
    """
    target_params = target
    origin_params = reference
    _check_voltage_point_input(target, elements)
    if origin_params is not None:
        _check_voltage_point_input(origin_params, elements)
    if do_align:
        qua.align(*elements)
    for element in elements:
        if origin_params is not None:
            amplitude = target_params[element].qua - origin_params[element].qua
        else:
            amplitude = target_params[element].qua
        kwargs = {
            'pulse': operation*qua.amp(amplitude),
            'element': element
            }
        if duration is not None:
            kwargs['duration'] = duration.qua
        logging.debug(
            "Arbok_go: Moving %s from %s to %s by %s", 
            element, reference, target, amplitude
            )
        if not isinstance(amplitude, (float, int)):
            qua.play(**kwargs)
        elif math.isclose(amplitude, 0, abs_tol= no_play_tolerance):
            logging.debug(
                "Arbok_go: Omitting %s since amplitude %s is small (th = %s)", 
                element, amplitude, no_play_tolerance
                )
        else:
            qua.play(**kwargs)
    if do_align:
        qua.align(*elements)

def _check_voltage_point_input(
        parameter_maps: ParameterMap[str, Voltage],
        elements: list[str]
        ) -> None:
    """
    Checks if input is of type list or string and returns input as list
    """
    if parameter_maps is None:
        print('Nothing to return!')
        return
    if not isinstance(parameter_maps, ParameterMap):
        raise TypeError(
            f"Must be of type ParameterMap. Is {type(parameter_maps)}")
    if not all([isinstance(p, Voltage) for _, p in parameter_maps.items()]):
        raise TypeError(
            "Not all parameters in the given ParameterMap are of type Voltage"
        )
    if not set(elements).issubset(parameter_maps):
        missing = set(elements) - parameter_maps.keys()
        raise ValueError(
            f"Missing elements in arbok.ramp parameter maps: {missing}")



"""Module containing generic sequence for a simple square pulse sequence"""
from dataclasses import dataclass
from qm import qua
from arbok_driver import arbok, ParameterClass, SubSequence
from arbok_driver.parameter_types import (
    List, Time, ParameterMap, Voltage)

@dataclass(frozen = True)
class SquarePulseParameters(ParameterClass):
    sticky_elements: List
    t_ramp: Time
    t_square_pulse: Time
    v_home: ParameterMap[str, Voltage]
    v_square: ParameterMap[str, Voltage]

class SquarePulseScalable(SubSequence):
    """
    Class containing parameters and sequence for a simple square pulse
    """
    PARAMETER_CLASS = SquarePulseParameters
    arbok_params: SquarePulseParameters
    def qua_sequence(self):
        """Macro that will be played within the qua.program() context"""
        qua.align(*self.arbok_params.sticky_elements.qua)
        arbok.ramp(
            elements= self.arbok_params.sticky_elements.qua,
            reference = self.arbok_params.v_home,
            target = self.arbok_params.v_square,
            duration = self.arbok_params.t_ramp,
            operation = 'unit_ramp',
            )
        qua.wait(
            self.arbok_params.t_square_pulse.qua,
            *self.arbok_params.sticky_elements.qua
            )
        arbok.ramp(
            elements= self.arbok_params.sticky_elements.qua,
            reference = self.arbok_params.v_square,
            target = self.arbok_params.v_home,
            duration = self.arbok_params.t_ramp,
            operation = 'unit_ramp',
            )



"""Init module loading example configurations"""
from .hardware_configs.divider_config import divider_config
from .hardware_configs.opx1000_config import opx1000_config
from .sequence_configs.coulomb_peaks_config import coulomb_peaks_set1_set2_conf
from .sequence_configs.parity_init_conf import parity_init_conf
from .sequence_configs.parity_readout_conf import parity_read_conf
from .sequence_configs.square_pulse_confs import (
    square_pulse_conf,
    square_pulse_scalable_conf,
)
from .sequence_configs.stability_map_config import stability_map_config
from .sequence_configs.device_config import device_config




# Single QUA script generated at 2026-03-28 03:07:58.818536
# QUA library version: 1.2.5


from qm import CompilerOptionArguments
from qm.qua import *

with program() as prog:
    v1 = declare(int, value=0)
    with infinite_loop_():
        pause()
        assign(v1, 0)
        align()
        play("ramp"*amp(0.5), "gate_1", duration=20)
        wait(50, "gate_1")
        play("ramp"*amp(0.5), "gate_1", duration=20)
        align()
        assign(v1, (v1+1))
        r0 = declare_stream()
        save(v1, r0)
        align()
    with stream_processing():
        r0.buffer(1).save("qm_driver_mock_measurement_shots")

config = None

loaded_config = None




"""Testing that the a sub-sequences parameter class is added correctly"""
import pytest
from arbok_driver import SubSequence, EmptyParameterClass
from arbok_driver.parameter_types import (
    Int, Voltage, ParameterMap
)
from .conftest import UserSubSequence

def test_missing_parameter_class() -> None:
    
    class SubSequenceNoParameterClass(SubSequence):
        pass
    assert not hasattr(SubSequenceNoParameterClass, 'arbok_params')
    assert SubSequenceNoParameterClass.PARAMETER_CLASS == EmptyParameterClass

def test_missing_parameter(mock_measurement) -> None:
    config_1 = {
        'par2': {'type': Int, 'value': int(10)},
        'par3': {'type': Voltage, 'value': 1.1},
        'v_home': {'type': Voltage, 'elements': {'P1': 5, 'J1': 6, }},
    }
    with pytest.raises(TypeError):
        _ = UserSubSequence(
            mock_measurement, 'sub_seq1', config_1)

def test_parameter_mapping_on_sub_sequence(empty_sub_seq_1) -> None:
    arbok_params = empty_sub_seq_1.arbok_params
    print(arbok_params.__dict__)
    assert len(arbok_params.__dict__) == 4
    assert hasattr(arbok_params, 'par1')
    assert isinstance(arbok_params.v_home, ParameterMap)
    assert not hasattr(arbok_params, 'par3')
    assert 'par3' in empty_sub_seq_1.parameters

def test_parameter_mapping_on_read_sequence(read_sequence_no_readouts) -> None:
    arbok_params = read_sequence_no_readouts.arbok_params
    print(arbok_params.__dict__)
    assert len(arbok_params.__dict__) == 4
    assert hasattr(arbok_params, 'par1')
    assert isinstance(arbok_params.v_home, ParameterMap)
    assert arbok_params.v_home["P1"].get() == 5
    assert not hasattr(arbok_params, 'par3')
    assert 'par3' in read_sequence_no_readouts.parameters



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
sys.path.insert(0, os.path.abspath("..")) 

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'arbok_driver'
copyright = '2025, Andreas Nickl'
author = 'Andreas Nickl'
release = 'v2.1.5'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_nb",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",       # Google/NumPy-style docstrings
    "sphinx_autodoc_typehints",  # show type hints in docs
]


templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
jupyter_execute_notebooks = "off"
autodoc_typehints = "signature"  # default is 'description'
napoleon_google_docstring = True
napoleon_numpy_docstring = False
autodoc_member_order = "bysource"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']



"""Module containing the MeasurementRunner class."""
from __future__ import annotations
from typing import TYPE_CHECKING
from datetime import datetime
import logging
import os
from pathlib import Path

import qcodes
from qm import generate_qua_script
import numpy as np
from qcodes.dataset.sqlite.database import get_DB_location

from .measurement_runner_base import MeasurementRunnerBase

if TYPE_CHECKING:
    from arbok_driver.measurement import Measurement
    from arbok_driver.measurement_runners.parameters.sequence_parameter import SequenceParameter
    from qm.program import Program
    from qcodes.dataset.data_set import DataSet as QcDataSet
    from qcodes.dataset.measurements import Measurement as QcMeasurement

class QCodesMeasurementRunner(MeasurementRunnerBase):
    """
    Helper class constructing QCoDeS measurement loops
    """

    def __init__(
        self,
        measurement: Measurement,
        ext_sweep_list: list[dict[SequenceParameter, np.ndarray]] | None = None,
        register_all: bool = False
        ):
        super().__init__(
            measurement=measurement,
            ext_sweep_list=ext_sweep_list,
            register_all=register_all
        )
        self.qc_measurement  = measurement.get_qc_measurement()
        self.datasaver: QcMeasurement | None = None
        self.qc_dataset: QcDataSet | None = None

    def _prepare_measurement(self) -> None:
        """
        Prepare the measurement for running.
            1) preparing parameters and gettables
            2) auto saving qua program
        """
        self._register_params_and_gettables()

    def _handle_keyboard_interrupt(self) -> None:
        pass

    def _wrap_up_measurement(self) -> None:
        """
        Wrap up the measurement after running.
            1) Fetches data as xarray dataset
            2) Sets the run id in the measurement
            3) Saves the QUA program as metadata to qcodes dataset
        """
        self.qc_dataset = self.datasaver.dataset
        self.measurement._set_dataset(self.qc_dataset.to_xarray_dataset())
        self.measurement._set_run_id(self.run_id)
        self._save_qua_program_as_metadata(
            self.measurement.qua_program,
            self.measurement.driver.device.config
            )

    def _run_measurement(self) -> QcMeasurement:
        """
        Builds the QCoDeS measurement object for the measurement.
        """
        with self.qc_measurement.run() as datasaver:
            self.datasaver = datasaver
            self.run_id = self.datasaver.run_id
            self._auto_save_qua_program()
            self._create_recursive_measurement_loop(self.ext_sweep_list)
            print("Measurement finished!")
        return datasaver

    def _save_results(self, results_xr: xr.Dataset) -> None:
        """
        Saves the results of the measurement to the datasaver.
        
        Args:
            datasaver (DataSaver): The QCoDeS DataSaver object to save results to.
        """
        ext_coords = {p.register_name: v for p, v in self.external_param_values.items()}
        results_xr = results_xr.sel(**ext_coords)
        ### Save gettables independently. This ensures correct shape handling
        ### for dependent parameters (gettable.unpack_self())
        for gettable in self.measurement.gettables.values():
            self.datasaver.add_result(
                *[(p, v) for p, v in self.external_param_values.items()],
                (gettable, results_xr[gettable.register_name].to_numpy())
                # (gettable, )
            )
        logging.debug("Results saved")

    def _register_params_and_gettables(self):
        """
        Register varied parameters and gettables for the measurement.
        """
        for param, _ in self.external_param_values.items():
            logging.debug(
                "Registering sequence parameter %s", param.full_name)
            self.qc_measurement.register_parameter(param)
        for _, gettable in self.measurement.gettables.items():
            logging.debug("Registering gettable %s", gettable.register_name)
            self.qc_measurement.register_parameter(
                gettable, setpoints = list(self.external_param_values.keys()))

    def _auto_save_qua_program(self) -> None:
        """
        Automatically saves the QUA program in a folder next to the database
        if the folder 'qua_programs' does not exist it is created
        """
        db_path = Path(get_DB_location())
        db_name = db_path.stem
        db_dir = db_path.parent
        programs_dir = Path(db_dir).resolve() / f"qua_programs__{db_name}/"
        programs_dir.mkdir(parents=True, exist_ok=True)
        save_path = programs_dir / f"{self.run_id}.py"
        print(f"Auto saving qua program next to database in {programs_dir}")
        save_path.write_text(
            generate_qua_script(
                self.measurement.qua_program,
                self.measurement.driver.device.config
                ),
            encoding="utf-8"
        )

    def _save_qua_program_as_metadata(
            self,
            qua_program: Program,
            opx_config: dict) -> None:
        """
        Saves the QUA program as metadata in the dataset.
        """
        if self.qc_dataset is not None:
            self.qc_dataset.add_metadata(
                "qua_program",
                generate_qua_script(qua_program, opx_config)
                )
        else:
            raise ValueError(
                "Dataset is not available yet for saving QUA program metadata."
                " Run measurement first."
                )



"""Module testing generic tuning interface"""

import pytest
import re
import numpy as np

from arbok_driver.examples.configurations import (
    parity_init_conf, parity_read_conf
)

def test_tuning_interface_requires_sweeps_and_gettables(mock_measurement) -> None:
    mock_measurement.driver.is_mock = True
    mock_measurement.mock_steps = 0
    mock_measurement.add_subsequences_from_dict({
        'paraity_init_even': {'config': parity_init_conf},
        'parity_read_even': {'config': parity_read_conf},
        'parity_init_odd': {'config': parity_init_conf},
        'parity_read_odd': {'config': parity_read_conf}
    })

    parameter_dicts = {
        't_ramp_init': {
            'qua_vars': {
                mock_measurement.parity_init_odd.arbok_params.t_ramp_over_crossing: 1
            },
            'bounds': (25, 1e4)
        }
    }

    with pytest.raises(ValueError):
        mock_measurement.initialize_tuning_interface(
            parameter_dicts=parameter_dicts,
            verbose=True
        )

    mock_measurement.set_sweeps({mock_measurement.iteration: np.arange(10)})

    with pytest.raises(ValueError):
        mock_measurement.initialize_tuning_interface(
            parameter_dicts=parameter_dicts,
            verbose=True
        )

def test_generic_tuning_interface_init(mock_measurement) -> None:
    mock_measurement.driver.is_mock = True
    mock_measurement.mock_steps = 0
    mock_measurement.add_subsequences_from_dict({
        'paraity_init_even': {'config': parity_init_conf},
        'parity_read_even': {'config': parity_read_conf},
        'parity_init_odd': {'config': parity_init_conf},
        'parity_read_odd': {'config': parity_read_conf}
    })
    mock_measurement.set_sweeps({mock_measurement.iteration: np.arange(100)})
    mock_measurement.register_gettables()
    _ = mock_measurement.get_qua_program_as_str()

    parameter_dicts = {
        't_ramp_init': {
            'qua_vars': {
                mock_measurement.parity_init_odd.arbok_params.t_ramp_over_crossing: 1
            },
            'bounds': (25, 1e4)
        },
        'read_detuning': {
            'qua_vars': {
                mock_measurement.parity_read_even.arbok_params.v_read['P1']: 1,
                mock_measurement.parity_read_even.arbok_params.v_read['P2']: -1,
                mock_measurement.parity_read_odd.arbok_params.v_read['P1']: 1,
                mock_measurement.parity_read_odd.arbok_params.v_read['P2']: -1,
            },
            'bounds': (-0.01, 0.01)
        },
    }

    tuning_interface = mock_measurement.initialize_tuning_interface(
        parameter_dicts=parameter_dicts,
        verbose=True
    )

    qua_prog_input_streams_str = mock_measurement.get_qua_program_as_str(recompile=True)
    assert len(re.findall('advance_input_stream', qua_prog_input_streams_str)) == 2

    def cost_strategy(gettables: dict) -> float:
        even = np.mean(gettables[mock_measurement.parity_read_even.p1p2.diff__p1p2])
        odd = np.mean(gettables[mock_measurement.parity_read_odd.p1p2.diff__p1p2])
        return float(even - odd)

    step_dict = {
        mock_measurement.parity_init_odd.arbok_params.t_ramp_over_crossing: 100,
        mock_measurement.parity_read_even.arbok_params.v_read['P1']: 0.1
    }

    gettable_results, saved_params = tuning_interface.run_parameter_set(step_dict)
    assert isinstance(gettable_results, dict)
    assert isinstance(saved_params, dict)
    assert len(gettable_results) > 0
    assert len(saved_params) > 0

    cost = cost_strategy(gettable_results)
    assert isinstance(cost, float)

    dataset = tuning_interface.run_cross_entropy_sampler(
        populations=[10, 10],
        cost_strategy=cost_strategy,
        select_frac=0.5,
        plot_histograms=True,
        sampling_params_to_plot=[('t_ramp_init', 'read_detuning')]
    )

    assert dataset is not None
    expected_min_keys = len(mock_measurement.gettables) + len(parameter_dicts) + 1
    assert len(dataset.data_vars) >= expected_min_keys
    assert 'rewards' in dataset.data_vars



"""Module containing test for Experiment class"""
from arbok_driver import Measurement
from arbok_driver.examples.experiments import (
    SquarePulseExperiment,
    SquarePulseExperimentWithDefaults
)
from arbok_driver.examples.configurations import square_pulse_conf

def test_measurement_creation_from_experiment(arbok_driver, dummy_device):

    experiment = SquarePulseExperiment()
    measurement = arbok_driver.create_measurement_from_experiment(
        experiment, 'qc_measurement')
    assert isinstance(measurement, Measurement)
    assert len(measurement.sub_sequences) == 1

def test_measurement_creation_with_defaults(arbok_driver, dummy_device):
    """
    Tests whether a experiment configuration is loaded corretly from the measurement defaults
    """
    experiment_default = SquarePulseExperimentWithDefaults()
    measurement_default = arbok_driver.create_measurement_from_experiment(
        experiment_default, 'qc_measurement')
    assert isinstance(measurement_default, Measurement)
    assert len(measurement_default.sub_sequences) == 1

def test_measurement_creation_overwrite(arbok_driver, dummy_device):
    experiment_overload = SquarePulseExperimentWithDefaults(
        square_pulse_config = {'config': square_pulse_conf})
    measurement_overload = arbok_driver.create_measurement_from_experiment(
        experiment_overload, 'qc_measurement')
    assert isinstance(measurement_overload, Measurement)
    assert len(measurement_overload.sub_sequences) == 1




# Single QUA script generated at 2024-06-19 16:02:16.257184
# QUA library version: 1.1.7

from qm.qua import *

with program() as prog:
    v1 = declare(int, value=0)
    v2 = declare(fixed, )
    v3 = declare(fixed, )
    v4 = declare(fixed, )
    v5 = declare(fixed, )
    v6 = declare(fixed, )
    v7 = declare(fixed, )
    v8 = declare(fixed, )
    with infinite_loop_():
        pause()
        align()
        measure("measure", "readout_element", None, integration.full("x", v2, ""), integration.full("y", v3, ""))
        assign(v4, (v2+v3))
        r2 = declare_stream()
        save(v2, r2)
        r3 = declare_stream()
        save(v3, r3)
        r4 = declare_stream()
        save(v4, r4)
        wait(50, )
        measure("measure", "readout_element", None, integration.full("x", v5, ""), integration.full("y", v6, ""))
        assign(v7, (v5+v6))
        r5 = declare_stream()
        save(v5, r5)
        r6 = declare_stream()
        save(v6, r6)
        r7 = declare_stream()
        save(v7, r7)
        assign(v8, (v4-v7))
        r8 = declare_stream()
        save(v8, r8)
        assign(v1, (v1+1))
        r1 = declare_stream()
        save(v1, r1)
    with stream_processing():
        r1.buffer(1).save("dummy_squence_shots")
        r2.buffer(1).save("dummy_readout__qubit1__reference__sensor1_I_buffer")
        r3.buffer(1).save("dummy_readout__qubit1__reference__sensor1_Q_buffer")
        r4.buffer(1).save("dummy_readout__qubit1__reference__sensor1_IQ_buffer")
        r5.buffer(1).save("dummy_readout__qubit1__read__sensor1_I_buffer")
        r6.buffer(1).save("dummy_readout__qubit1__read__sensor1_Q_buffer")
        r7.buffer(1).save("dummy_readout__qubit1__read__sensor1_IQ_buffer")
        r8.buffer(1).save("dummy_readout__qubit1__diff_buffer")


config = None

loaded_config = None




"""Module containing fixtures for pytests"""
from dataclasses import dataclass
import pytest 

from arbok_driver import (
    ArbokDriver,
    Device,
    Measurement,
    ParameterClass,
    ReadSequence,
    SequenceParameter,
    SubSequence,
)
from arbok_driver.examples.configurations import (
    device_config,
    square_pulse_conf,
    square_pulse_scalable_conf,
    opx1000_config,
    divider_config,
    stability_map_config
)
from arbok_driver.examples.sub_sequences import (
    SquarePulse,
    SquarePulseScalable,
    StabilityMap
)
from arbok_driver.parameter_types import (
    Amplitude, Int, ParameterMap,  Voltage
)

@dataclass(frozen = True)
class SubSequenceParameterClass(ParameterClass):
    par1: Amplitude
    par2: Amplitude
    iteration: Int
    v_home: ParameterMap[str, Voltage]

@dataclass(frozen = True)
class SubSequenceParameterClassAlt(ParameterClass):
    par4: Amplitude
    par5: Amplitude
    iteration: Int
    v_home: ParameterMap[str, Voltage]

class UserSubSequence(SubSequence):
    PARAMETER_CLASS = SubSequenceParameterClass

class UserSubSequenceAlt(SubSequence):
    PARAMETER_CLASS = SubSequenceParameterClassAlt

class UserReadSequence(ReadSequence):
    PARAMETER_CLASS = SubSequenceParameterClass

@pytest.fixture
def dummy_device():
    """Returns dummy device instance"""
    return Device(
        name = 'dummy_device',
        opx_config = opx1000_config,
        divider_config = divider_config,
        master_config = device_config
    )

@pytest.fixture
def arbok_driver(dummy_device) -> ArbokDriver: # type: ignore[reportInvalidTypeForm]
    """Returns ArbokDriver instance"""
    driver = ArbokDriver(name = 'arbok_driver', device = dummy_device)
    yield driver
    driver.close()

@pytest.fixture
def mock_measurement(arbok_driver, dummy_device) -> Measurement: # type: ignore[reportInvalidTypeForm]
    """Returns dummy measurement instance"""
    measurement = Measurement(
        parent = arbok_driver,
        name = 'mock_measurement'
        )
    yield measurement
    arbok_driver.submodules.pop(measurement.full_name)
    arbok_driver.measurements.remove(measurement)
    del measurement

@pytest.fixture
def empty_sub_seq_1(mock_measurement, dummy_device):
    """Returns dummy subsequence with a few parameters"""
    config_1 = {
        'par1': {'type': Int, 'value': int(1)},
        'par2': {'type': Int, 'value': int(10)},
        'par3': {'type': Voltage, 'value': 1.1},
        'v_home': {'type': Voltage, 'elements': {'P1': 5, 'J1': 6, }},
    }
    seq1 = UserSubSequence(mock_measurement, 'sub_seq1', config_1)
    yield seq1

@pytest.fixture
def empty_sub_seq_2(mock_measurement, dummy_device):
    """Returns dummy subsequence with a few parameters"""
    config_2 = {
        'par4': {'type': Int, 'value': int(2)},
        'par5': {'type': Int, 'value': int(3)},
        'v_home': {'type': Voltage, 'elements': {'P1': 0, 'J1': 7, }},
    }
    seq2 = UserSubSequenceAlt(
        mock_measurement, 'sub_seq2', config_2)
    yield seq2

@pytest.fixture
def read_sequence_no_readouts(mock_measurement, dummy_device):
    """Returns dummy subsequence with a few parameters"""
    config_1 = {
        'parameters': {
            'par1': {'type': Int, 'value': int(1)},
            'par2': {'type': Int, 'value': int(10)},
            'par3': {'type': Voltage, 'value': 1.1},
            'v_home': {'type': Voltage, 'elements': {'P1': 5, 'J1': 6, }},
        },
        'signals': [],
        'readout_groups': {}
    }
    seq1 = UserReadSequence(
        mock_measurement, 'read_seq', config_1)
    yield seq1

@pytest.fixture
def stability_map(mock_measurement, dummy_device):
    """Returns dummy subsequence with a few parameters"""
    stability_map = StabilityMap(
        mock_measurement, 'stability_map', stability_map_config)
    yield stability_map

@pytest.fixture
def measurement_parameter(mock_measurement):
    mock_measurement.add_parameter(
        parameter_class = SequenceParameter,
        name  = 'dummy_param',
        initial_value = 0,
        scale = 1,
        get_cmd = None,
        set_cmd = None,
    )
    mock_measurement.dummy_param.qua_var = 'placeholder'
    return mock_measurement.dummy_param

@pytest.fixture
def square_pulse(mock_measurement):
    square_pulse = SquarePulse(
        mock_measurement, "square_pulse", square_pulse_conf)
    yield square_pulse

@pytest.fixture
def square_pulse_scalable(mock_measurement) -> SquarePulseScalable:
    square_pulse = SquarePulseScalable(
        mock_measurement, "square_pulse", square_pulse_scalable_conf
        )
    yield square_pulse



""" Module containing SequenceParameter class """
from __future__ import annotations
from typing import TypeVar, Generic, Optional
import logging

from numpy import ndarray
import numpy as np
from qcodes.parameters import Parameter
from qm import qua

T_co = TypeVar("T_co", covariant=True)

class SequenceParameter(Parameter, Generic[T_co]):
    """
    A parameter wrapper that adds the respective element as attribute

    TODO: Write get_raw abstract method without crashing sequence compilation
    """

    validator = None
    sweep_validator = None
    unit = ''
    qua_type = int
    var_type = None
    input_stream = None
    qua_sweeped = False
    qua_sweep_arr = None
    qua_var = None
    value = None
    can_be_parameterized = False
    scale = 1

    def __init__(self, *args, var_type = None, element = None, **kwargs):
        """
        Constructor for 'SequenceParameter' class

        Args:
            elements (list): Elements that should be influenced by parameter
            config_name (str): Name of the parameter in the sequence config dict
                essentially name without the element
        """
        try:
            super().__init__(*args, **kwargs)
        except Exception as e:
            raise ValueError(
                f"Error initializing SequenceParameter: {kwargs['name']}: {e}"
                ) from e
        self.element = element
        if var_type is not None:
            self.var_type = var_type

    @property
    def sequence_path(self) -> str:
        """Returns the path through all parent sequences above"""
        return f"{self.instrument.get_sequence_path()}_{self.name}"

    @property
    def full_name(self) -> str:
        """Returns the full name of the parameter"""
        return self.sequence_path

    @property
    def qua(self) -> T_co:
        """Getter method for parameter"""
        if self.qua_var is not None:
            return self.qua_var
        else:
            return_value = self.call_method()
            return return_value

    def convert_to_real_units(self, value):
        """
        Converts the value of the parameter to real units

        Args:
            value (float|int): Value to be converted

        Returns:
            float|int: Converted value
        """
        return value

    def __call__(
            self, value: Optional[float | int | ndarray] = None,
            **kwargs
            ) -> T_co | None:
        return self.call_method(value, **kwargs)

    def call_method(
            self,
            value: Optional[float | int | ndarray] = None,
            **kwargs
            ) -> T_co | None:
        """
        Method being executed when SequenceParameter is called.
        
        Args:
            value (Optional[float | int]): Value if given sets

        Returns:
            float|int|np.ndarray: Parameter value if no input value is given
        """
        if value is None:
            if self.qua_var is not None:
                return self.qua_var
            else:
                return_value = self.get_raw()
                return return_value
        else:
            if self.qua_var is not None:
                raise ValueError(
                    f"Parameter {self.full_name}'s qua_var is not None so it"
                    "cant be set. It is likely being swept or used as a"
                    "variable elsewhere!"
                )
            self.set(value)

    def reset(self) -> None:
        """
        In case we have switched to a sweep state, reset the validator to a scalar.
        """
        while self.remove_validator():
            pass
        self.add_validator(self.validator)

    def qua_declare(self, setpoints):
        """
        Declares the parameter inside qua code as variable and sets its class
        attributes accordingly. Note: This method can only be called inside the
        qua.program() context manager

        Args:
            setpoints (list, numpy.array): Setpoints for parameter sweep
        """
        logging.debug(
            "Adding qua %s variable for %s on subsequence %s (stream %s)",
            type(self.get()), self.name, self.instrument.name, self.input_stream
        )
        if self.var_type == int:
            setpoints = np.array(setpoints, dtype = int)
        else: 
            setpoints = np.array(setpoints)
        self.qua_sweeped = True

        self.qua_var = qua.declare(self.var_type)
        if self.can_be_parameterized:
            pass
        elif self.input_stream is None:
            self.qua_sweep_arr = qua.declare(
                self.var_type, value = setpoints*self.scale)
        else:
            self.input_stream = qua.declare_input_stream(
                t = self.var_type,
                name = self.sequence_path,
                size = int(setpoints)
            )

    def add_stream_param_to_sequence(self):
        """Adds input stream to sequence"""
        if self.input_stream is not None:
            sequence = self.instrument.measurement
            sequence.add_input_stream_parameter(self)
        else:
            raise ValueError(
                f"Parameter {self.name} has no input stream to add"
            )



from .coulomb_peaks import CoulombPeaks
from .parity_initialization import ParityInit
from .parity_readout import ParityRead
from .square_pulse import SquarePulse
from .square_pulse_legacy import SquarePulseLegacy
from .square_pulse_scalable import SquarePulseScalable
from .stability_map import StabilityMap


""" Module containing BaseSequence class """
from __future__ import annotations
from abc import ABC
from typing import TYPE_CHECKING, Optional
import copy
import types
import warnings
import logging
from functools import reduce
from anytree import RenderTree
from qcodes.instrument import InstrumentModule

from qm import SimulationConfig, generate_qua_script, qua, QuantumMachinesManager
from qm.simulate.credentials import create_credentials

from .parameters.sequence_parameter import SequenceParameter
from .parameter_class import ParameterClass, EmptyParameterClass
from .parameter_types import ParameterMap
from . import utils

if TYPE_CHECKING:
    from qm.grpc.qua import QuaProgram

    from .device import Device
    from .sub_sequence import SubSequence

class SequenceBase(InstrumentModule, ABC):
    """
    Class describing a subsequence of a QUA programm (e.g Init, Control, Read). 
    """
    PARAMETER_CLASS: type[ParameterClass]
    _enforce_parameter_class: bool = False
    def __init__(
            self,
            parent,
            name: str,
            sequence_config: Optional[dict | None] = None,
            check_step_requirements: Optional[bool] = False,
            **kwargs
            ):
        """
        Constructor class for `SequenceBase` class
        
        Args:
            name (str): Name of the program
            sequence_config (dict): Dictionary containing all device parameters
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(parent, name, **kwargs)
        self.parent.add_submodule(self.name, self)
        setattr(self.parent, self.short_name, self)

        self.device: Device = self.parent.device
        self.elements = self.device.elements
        self.sequence_config = sequence_config
        self.check_step_requirements = check_step_requirements

        self._sub_sequences = []
        self._sub_sequence_dict = {}
        self._gettables = []
        self._parameter_maps: dict = {}
        self._qua_program_as_str = None
        self.add_qc_params_from_config(self.sequence_config)

    def __init_subclass__(cls, **kwargs) -> None:
        """Enforces child classes to define PARAMETER_CLASS class attribute"""
        super().__init_subclass__(**kwargs)
        if cls._enforce_parameter_class:
            if 'PARAMETER_CLASS' not in cls.__dict__:
                cls.PARAMETER_CLASS = EmptyParameterClass
        cls._enforce_parameter_class = True

    @classmethod
    def config_template(cls)  -> None:
        """
        The user can get an example config template.
        This feature is useful if building from scratch or for UI 
        prompting.
            
        Returns:
            A dictionary with an example config template. 
        """
        return {}

    def qua_declare(self) -> None:
        """Contains raw QUA code to initialize the qua variables"""
        for sub_sequence in self.sub_sequences:
            sub_sequence.qua_declare()

    def qua_before_sweep(self) -> None:
        """Contains raw QUA code that is being executed before sweeps"""
        for sub_sequence in self.sub_sequences:
            sub_sequence.qua_before_sweep()

    def qua_before_sequence(self) -> None:
        """Contains raw QUA code that is being executed before the sequence"""
        for sub_sequence in self.sub_sequences:
            sub_sequence.qua_before_sequence()

    def qua_sequence(self) -> None:
        """Contains raw QUA code to define the pulse sequence"""
        for sub_sequence in self.sub_sequences:
            sub_sequence.qua_sequence()

    def qua_after_sequence(self) -> None:
        """Contains raw QUA code that is being executed after the sequence"""
        for sub_sequence in self.sub_sequences:
            sub_sequence.qua_after_sequence()

    def qua_stream(self) -> None:
        """Contains raw QUA code to define streams"""
        for sub_sequence in self.sub_sequences:
            sub_sequence.qua_stream()

    @property
    def sub_sequences(self) -> list[SubSequence]:
        """
        List of `SubSequences`s that build the given sequence
        """
        return self._sub_sequences

    @property
    def sub_sequence_dict(self) -> dict:
        """
        List of `SubSequences`s that build the given sequence
        """
        structure_dict = {}
        for sub_sequence in self._sub_sequences:
            if len(sub_sequence.sub_sequences) > 0:
                sub_dict = sub_sequence.sub_sequence_dict
                structure_dict[sub_sequence.short_name] = sub_dict
            else:
                structure_dict[sub_sequence.short_name] = {}
        return structure_dict # {self.short_name: structure_dict}

    @property
    def gettables(self) -> list:
        """
        List of `GettableParameter`s that can be registered for acquisition
        in a `program`
        """
        return self._gettables

    def add_subsequence(self, new_sequence) -> None:
        """Adds a subsequence to self"""
        self._sub_sequences.append(new_sequence)

    def get_parameters_and_maps(
            self, param_names: list[str]
            ) -> dict[str, SequenceParameter | ParameterMap]:
        param_dict = {}
        for param_name in param_names:
            if param_name in self.parameters:
                param_dict[param_name] = self.parameters[param_name]
            elif param_name in self._parameter_maps:
                param_dict[param_name] = self._parameter_maps[param_name]
        return param_dict

    def add_qc_params_from_config(self, config):
        """ 
        Creates QCoDeS parameters for all entries of the config 
        TODO: Use custom Parameter types for times -> setting in ns ! (cycles)
                use validator to check if ns are a multiple of 4 (1 cycle)
        TODO: if voltage add scale = 0.5 and validate if |v| <= 0.5

        Args:
            config (dict): Configuration containing all sequence parameters
        """
        if config is None:
            logging.info("No params added to %s (no sequence_config)", self.name)
            return
        elif not isinstance(config, dict):
            raise ValueError(
                f"Conf for {self.name} must be of type dict, is {type(config)}")
        if 'parameters' in config:
            config = config['parameters']
            logging.info(
                "Set param subset of given conf file as param conf for in %s",
                self.name
                )
        for param_name, param_dict in config.items():
            self._add_param(param_name, param_dict)

    def draw_sub_sequence_tree(self) -> None:
        """
        Draws a tree of the subsequences of the sequence with their names and
        types
        """
        root_node = utils.dict_to_anytree(self.short_name, self.sub_sequence_dict)
        for pre, _, node in RenderTree(root_node):
            print(f"{pre}{node.name}")

    def get_qua_program_as_str(self, recompile: bool = False) -> str:
        """Returns the qua program as str. Will be compiled if it wasnt yet"""
        if self._qua_program_as_str is None or recompile:
            self.get_qua_program()
        return self._qua_program_as_str

    def get_qua_program(self, simulate = False, config = None) -> QuaProgram:
        """
        Composes the entire sequence by searching recursively through init, 
        sequence and stream methods of all subsequences and their subsequences.
        The respective qua sequence will only be added once the recursive
        scans have reached the lowest level of sequences (e.g sequences have no
        sub-sequences anymore)

        Args:
            simulate (bool): Flag whether program is simulated
        Returns:
            program: Program compiled into QUA language
        """
        with qua.program() as prog:
            self.get_qua_code(simulate)
        self._qua_program_as_str = generate_qua_script(prog, config)
        return prog

    def get_qua_code(self, simulate = False):
        """
        Composes the entire qua sequence in qua code. Only execurte with
        qm.qua.program() environment
        
        Args:
            simulate (bool): True if program is generated for simulation
        """
        if not hasattr(self, "measurement"):
            raise ReferenceError(
                "The sub sequence {self.name} is not linked to a measurement")
        self.qua_declare_sweep_vars()
        self.qua_declare()

        ### An infinite loop starting with a pause is defined to sync the QM
        ### with the client
        with qua.infinite_loop_():
            if not simulate:
                qua.pause()

            ### Check requirements are set to True if the sequence is simulated
            if simulate:
                for qua_var in self.measurement.step_requirements:
                    qua.assign(qua_var, True)
            ### The sequences are run in the order they were added
            ### Before_sweep methods are run before the sweep loop
            self.qua_before_sweep()
            # self.recursive_qua_generation(seq_type = 'before_sweep')

            ### qua_sequence methods of sub_sequences are called recursively
            ### If parent sequence is present the sweep generation is added
            if hasattr(self.measurement, 'sweeps'):
                self.recursive_sweep_generation(
                    self.measurement.sweeps)
            else:
                self.qua_sequence()

        ### Stream processing is added after the sequences
        with qua.stream_processing():
            self.qua_stream()

    def print_qua_program_to_file(self, file_name: str):
        """Creates file with 'filename' and prints the QUA code to this file"""
        with open(file_name, 'w', encoding="utf-8") as file:
            file.write(self.get_qua_program_as_str())

    def qua_declare_sweep_vars(self) -> None:
        """ Declares all sweep variables as QUA with their correct type """
        logging.debug("Start declaring QUA variables in %s", self.name)
        for sweep in self.measurement.sweeps:
            for param, setpoints in sweep.config.items():
                if isinstance(param, SequenceParameter):
                    logging.debug("Declaring %s as %s",
                                    param.name, param.var_type)
                    param.qua_declare(setpoints)
                else:
                    raise TypeError(
                        f"Parameter {param} is not of type SequenceParameter"
                    )

    def recursive_sweep_generation(self, sweeps):
        """
        Recursively generates QUA parameter sweeps by introducing one nested QUA
        loop per swept axis. The last given sweep and its corresponding
        setpoints are in the innermost loop.
        TODO: Reimplement a fast version of this for non-paired parameter 
            sweeps
        Args:
            sweeps (list): list of Sweep objects
        """
        if len(sweeps) == 0:
            ### this condition gets triggered if we arrive at the innermost loop
            self.qua_before_sequence()
            self.qua_sequence()
            self.qua_after_sequence()
            return
        new_sweeps = sweeps[1:]
        current_sweep = sweeps[0]
        logging.debug("Adding qua loop for %s",
            [par.name for par in current_sweep.parameters])

        # In some cases, the current sweep needs knowledge of the next sweep,
        # for example the snake variable needs to be reset each outer loop.
        next_sweep = None
        if len(new_sweeps):
            next_sweep = new_sweeps[0]
        current_sweep.qua_generate_parameter_sweep(
            lambda: self.recursive_sweep_generation(new_sweeps),
            next_sweep
            )
        return

    def reset(self) -> None:
        """
        On reset, ensure param validators are no longer sweep_validators
        """
        for k, param in self.parameters.items():
            if isinstance(param, SequenceParameter):
                self.parameters[k].reset()
        for sub in self.sub_sequences:
            sub.reset()

    def remove_subsequences(self) -> None:
        """Removes all subsequences from the sequence"""
        while len(self.sub_sequences) > 0:
            sub = self._sub_sequences.pop()
            sub.remove_subsequences()
            delattr(self, sub.short_name)
            if sub.short_name in globals():
                del globals()[sub.short_name]
        self._sub_sequences = []

    def _add_param(self, param_name: str, param_dict: dict) -> SequenceParameter | None:
        """
        Adds parameter based on the given parameter configuration
        
        Args:
            param_name (str): Name of the parameter
            param_dict (dict): Must contain 'unit' key and optionally 'value'
                or 'elements' for element wise defined parameters

        Returns:
            SequenceParameter: just created parameter
        """
        logging.debug("Adding %s to %s", param_name, self.name)
        param_dict = self._reshape_param_dict(param_name, param_dict)
        if 'elements' in param_dict:
            self._add_element_params(param_name, param_dict)
            return None
        self._check_param_dict(param_name, param_dict)
        # If the parameter already exists on another sub-sequence or measurement
        # we add the sequence name to the parameter name to avoid conflicts
        # To reference the parameter later, we also add the short name of the
        # sequence to the parameter name
        attr_exists_already = hasattr(self, param_name)
        if attr_exists_already:
            short_param_name = param_name
            param_name = f"{self.short_name}__{param_name}"
        new_param = self.add_parameter(
            name  = param_name,
            get_cmd = None,
            set_cmd = None,
            register_name = f"{self.short_name}__{param_name}",
            **param_dict
        )
        if attr_exists_already:
            setattr(self, short_param_name, new_param)
        return new_param

    def _reshape_param_dict(
            self, param_name: str, param_dict: dict
            ) -> dict:
        """
        Reshapes the parameter dict to fit the requirements of the respective
        'SequenceParameter' constructor and checks if the parameter dict is ok.

        Args:
            param_name (str): Name of the parameter
            param_dict (dict): Dictionary containing the parameter configuration
                to be reshaped

        Returns:
            dict: Reshaped parameter dict
        """
        if 'type' not in param_dict and 'parameter_class' not in param_dict:
            raise KeyError(
                f"Config for parameter '{param_name}' on '{self.full_name}'"
                " does not contain a 'type' key."
                " Please provide a type of the parameter to be added."
                " Find available ones in arbok_driver/parameter_types.py."
                " Or use a custom one.")
        if 'type' in param_dict:
            param_dict['parameter_class'] = param_dict['type']
            del param_dict['type']
        if 'value' in param_dict:
            param_dict['initial_value'] = param_dict['value']
            del param_dict['value']
        if 'label' not in param_dict:
            param_dict['label'] = param_name
        for key in ['unit', 'var_type', 'vals', 'scale']:
            if key not in param_dict:
                param_dict[key] = getattr(
                    param_dict['parameter_class'], key, None)
        return param_dict

    def _add_element_params(self, param_name: str, param_dict: dict):
        """
        Adds element wise parameter based on the given parameter configuration
        
        Args:
            param_name (str): Name of the parameter
            param_dict (dict): Must contain 'elements' key and optionally 'value'
                or 'elements' for element wise defined parameters
        """
        element_mapping = {}
        for element, value in param_dict['elements'].items():
            element_param_dict = {
                'element' : element,
                'value' : value,
                'label' : f"{element}: {param_dict['label']}",
                }
            if element in self.device.divider_config:
                scale = self.device.divider_config[element]['division']
                element_param_dict['scale'] = scale

            param_dict_copy = copy.deepcopy(param_dict)
            del param_dict_copy['label']
            param_dict_copy.update(element_param_dict)
            del param_dict_copy['elements']

            element_mapping[element] = self._add_param(
                param_name = f'{param_name}_{element}',
                param_dict = param_dict_copy
                )
        self._parameter_maps[param_name] = ParameterMap(element_mapping)

    def _check_param_dict(self, param_name: str, param_dict: dict) -> None:
        """
        Checks if the given parameter dict is valid. Raises an error if not.
        
        Args:
            param_dict (dict): Dictionary containing the parameter configuration
        """
        if 'parameter_class' not in param_dict:
            raise ValueError(
                f"Parameter {param_name} does not contain a 'parameter_class'")
        if 'label' not in param_dict:
            raise ValueError(
                f"Parameter {param_name} does not contain a 'label' key")
        if 'initial_value' not in param_dict and 'elements' not in param_dict:
            raise ValueError(
                f"Parameter {param_name} does not contain an 'value'"
                " key")
        if 'elements' in param_dict:
            raise KeyError(
                f"Config for parameter {param_name} contains"
                "'elements' key. Error in preparation. Check" \
                "`_add_element_params` method."
                )

    def run_remote_simulation(self, host, port, duration: int):
        """
        Simulates the MW sequence on a remote simulator on the host

        Args:
            host (str): Host address
            duration (int): Amount of cycles (4ns/cycle) to simulate

        Returns:
            SimulatedJob: QM job containing simulation results
        """
        qmm = QuantumMachinesManager(
            host = host,
            port = port,
            credentials=create_credentials()
        )
        simulated_job = qmm.simulate(
            self.device.config,
            self.get_qua_program(simulate = True),
            SimulationConfig(duration=duration)
            )

        devices = simulated_job.get_simulated_devices()
        utils.plot_qmm_simulation_results(devices)
        return simulated_job

    def find_parameters_from_keywords(
        self,
        keys: str | list,
        elements: list[str] | None = None
        ) -> dict:
        """
        Returns a list containing all parameters of the seqeunce with names that
        contain one of names in the 'keys' list.
        TODO:   - raise error if no params were found

        Args:
            keys (str | list): string with parameter name sub-string or list of
                those
            elements (list)
        Returns:
            Dict of parameters containing substrings from keys in their name
                with element as key
        """
        if elements is None:
            elements = list(self.device.elements)
        if isinstance(keys, str):
            keys = [keys]
        elif not isinstance(keys, list):
            raise ValueError(
                f"key has to be of type list or str, is {type(keys)}")
        element_dict = {element: {} for element in elements}
        for element in elements:
            for key in keys:
                param = getattr(self, f"{key}_{element}")
                element_dict[element][key] = param
        return element_dict

    def _add_subsequence(
        self,
        name: str,
        sequence_config: dict,
        namespace_to_add_to: dict | None= None,
        **kwargs
        ) -> SubSequence:
        """
        Adds a subsequence to the sequence
        
        Args:
            name (str): Name of the subsequence
            sequence_config (dict): Config containing all measurement params
            namespace_to_add_to (dict): Name space to insert the
                subsequence into (e.g locals(), globals()) defaults to None
        """
        if 'sequence' not in sequence_config:
            raise KeyError(
                f"The given config for {self.full_name}__{name} does not contain a "
                "'sequence' key with a subsequence type to configure for."
                )
        subsequence = sequence_config['sequence']
        if not issubclass(subsequence, SequenceBase):
            raise TypeError(
                "Subsequence must be of type SubSequence")
        seq_instance = subsequence(
            parent = self,
            name = name,
            sequence_config = sequence_config,
            **kwargs
            )
        setattr(self, name, seq_instance)
        if namespace_to_add_to is not None:
            name_space = namespace_to_add_to
            name_space[name] = seq_instance
        return seq_instance

    def _add_subsequences_from_dict(
            self,
            default_sequence,
            subsequence_dict: dict,
            namespace_to_add_to: dict | None = None) -> None:
        """
        Adds subsequences to the sequence from a given dictionary

        Args:
            default_sequence (SubSequence): Default subsequence to be used if
                no sequence is given in the subsequence_dict. Since is meant to
                be SubSequence which is a child of SequenceBase, therefore it
                cant be directly referenced here but is given from the
                respective child
            subsequence_dict (dict): Dictionary containing the subsequences
            namespace_to_add_to (dict): Name space to insert the
                subsequence into (e.g locals(), globals()) defaults to None
                    """
        if type(self) is SequenceBase:
            raise TypeError(
                "Method is meant to be used in SubSequence or Measurement!")
        if isinstance(subsequence_dict, types.SimpleNamespace):
            subsequence_dict = vars(subsequence_dict)
        for name, seq_conf  in subsequence_dict.items():
            ### Check whether a subsequence is configured or empty
            if seq_conf is None:
                raise ValueError(
                    f"Conf for {name} is None, please provide dict. E.g: "
                    r"{'sequence': SubSequence, 'config': {}}"
                    )
            if 'sub_sequences' in seq_conf:
                ### If empty SubSequence, create one deeper nesting layer
                if any(k in seq_conf.keys() for k in ['sequence', 'config']):
                    raise KeyError(
                        f"Subsequence {name} contains 'sub_sequences' key but "
                        "also 'sequence' or 'config' keys. If you give "
                        f"sub_sequences, {name} has to be empty."
                    )
                sequence_config = {
                    'sequence': default_sequence,
                    'parameters': {},
                    }
                kwargs = seq_conf.get('kwargs', {})
                seq_instance = self._add_subsequence(
                    name = name,
                    sequence_config = sequence_config,
                    namespace_to_add_to = namespace_to_add_to,
                    **kwargs
                    )
                seq_instance.add_subsequences_from_dict(
                    seq_conf["sub_sequences"], namespace_to_add_to)
            elif any(k in seq_conf.keys() for k in ['sequence', 'config']):
                self._prepare_adding_subsequence(
                    name, seq_conf, namespace_to_add_to)
            elif not seq_conf:
                pass
            else:
                raise KeyError(
                    f"Someting is wrong with the conf for {name}. "
                    f"Check the config {seq_conf}."
                    )

    def _prepare_adding_subsequence(
        self,
        name: str,
        seq_conf: dict,
        namespace_to_add_to: dict | None = None
        ) -> None:
        ### Check if config available and of type dict
        sub_seq_conf = {'parameters': {}}
        if 'config' in seq_conf:
            if not isinstance(sub_seq_conf, dict):
                raise ValueError(
                    f"Subsequence config ({name}) must be of type dict,"
                    f" is {type(sub_seq_conf)}")
            if 'parameters' not in seq_conf['config']:
                raise KeyError(
                    f"Config for {self.full_name}__{name} does not contain "
                    "'parameters' key."
                )
            if 'sequence' not in seq_conf['config']:
                raise ValueError(
                    f"Config for {self.full_name}__{name} does not specify "
                    "the SubSequence it configures. E.g 'sequence key is missing'"
                )
            sub_seq_conf = seq_conf['config']
        ### Check if kwargs available and of type dict
        if 'kwargs' in seq_conf:
            kwargs = seq_conf['kwargs']
            if not isinstance(kwargs, dict):
                raise ValueError(
                    f"Kwargs must be of type dict, is {type(kwargs)}")
        else:
            kwargs = {}
        ### Check if sequence is available and of type SequenceBase
        if 'sequence' in seq_conf:
            if 'sequence' in sub_seq_conf:
                warnings.warn(
                    "If both 'config' and 'sequence' are given, 'sequence' "
                    "will be used and the seq given in 'config' will be "
                    f"ignored. Sequence {name}: " 
                    f"{sub_seq_conf['sequence'].__name__} -> "
                    f"{seq_conf['sequence'].__name__}",
                )
            sub_seq_conf['sequence'] = seq_conf['sequence']
        _ = self._add_subsequence(
            name = name,
            sequence_config = sub_seq_conf,
            namespace_to_add_to = namespace_to_add_to,
            **kwargs
            )

    def find_parameters(self, key: str, elements: list | None = None) -> dict:
        """
        Finds all parameters generated from elements and a the given key.
        Similar to `find_parameters_from_keywords` but returns a non nested
        dict with elements as keys and SequenceParameters as key. This function
        gets its elements straigt from the given quantum machines config,
        therefore only params with elements that are known to the hardware are
        returned.

        Args:
            key (str): Name of the searched parameters
            elements (list): List of elements to be searched for

        Returns:
            dict: Dict with all found SequenceParameters and elements as keys
        """
        parameters = {}
        if elements is None:
            elements = self.device.elements
        for element in elements:
            if hasattr(self, f"{key}_{element}"):
                parameters[element] = getattr(self, f"{key}_{element}")
        return parameters

    def find_parameter(self, key: str, element: str):
        """Returns parameter with a certain key for a given element"""
        parameter = getattr(self, f"{key}_{element}")
        return parameter
                                                                                                                            
    def get_attribute_by_path(self, path):                                                                                                               
        """Access a nested attribute using a dot-separated string.
        Args:
            path: a dot delimited path to get the variable from.
        Returns:
            The variable matching the string path
        """                                                                                   
        return reduce(getattr, path.split('.'), self)     

    def find_parameter_from_str_path(
            self, path: str | list[str]) -> SequenceParameter:
        """
        Returns the parameter from the given path

        Args:
            path (str): Path to the parameter

        Returns:
            SequenceParameter: Parameter at the given path
        """
        if isinstance(path, str):
            path = path.split('__')
        if not isinstance(path, list):
            raise ValueError(
                f"path has to be of type list or str, is {type(path)}")
        if len(path) == 1:
            return getattr(self, path[0])
        else:
            return getattr(self, path[0]).find_parameter_from_str_path(path[1:])

    def ask_raw(self, *args):
        """Overwrites abstract method"""
        raise NotImplementedError("This driver does not support `ask_raw`")

    def set_raw(self, *args):
        """Overwrites abstract method"""
        raise NotImplementedError("This driver does not support `set_raw`")

    def set_params_with_unit_to_value(self, unit: str, value: any) -> None:
        """
        Sets all parameters with the given unit to the given value

        Args:
            unit (str): Unit of the parameters to be set
            value (any): Value to be set
        """
        for param_name, param in self.parameters.items():
            if param.unit == unit:
                print(f"Setting param {param_name} to {value}")
                logging.debug("Setting param %s to %s", param_name, value)
                param(value)



"""Module containing a minimal voltage divider config """
opx_scale = 2
divider_config = {
    'P1': {
        'division': 3*opx_scale,
    },
    'P2': {
        'division': 3*opx_scale,
    },
    'P3': {
        'division': 3*opx_scale,
    },
    'gate_2': {
        'division': 1*opx_scale,
    },
    'readout_element': {
        'division': 1*opx_scale
    }
}



"""Module containing threshold readout helper"""
from dataclasses import dataclass

from qm import qua
from arbok_driver import (
    AbstractReadout,
    ReadSequence,
    Signal,
    ParameterClass
)

from arbok_driver.parameter_types import Voltage

@dataclass(frozen=True)
class ThresholdParameters(ParameterClass):
    threshold: Voltage

class Threshold(AbstractReadout):
    """Helper class enabling threshold readout from a given signal"""
    PARAMETER_CLASS = ThresholdParameters
    arbok_params: ThresholdParameters

    def __init__(
        self,
        name: str,
        read_sequence: ReadSequence,
        signal: Signal,
        save_results: bool,
        parameters: dict,
        charge_readout: str,
        ):
        """
        Constructor method of `Threshold` readout helper
        
        Args:
            name (str): Name of readout
            read_sequence (arbok_driver.ReadSequence): ReadSequence performing the
                abstract readout
            signal (arbok_driver.Signal): Signal to be measured
            save_results (bool): Whether to save the results
            parameters (dict): Parameters to be added to the read sequence
            charge_readout (str): Name of the charge readout gettable
        """
        super().__init__(
            name = name,
            read_sequence = read_sequence,
            signal = signal,
            save_results = save_results,
            parameters = parameters
        )
        self.signal = signal
        self.charge_readout = charge_readout
        self._create_gettables()

    def qua_measure(self):
        """
        Measures the given gettables and assigns the result to the variables.
        Only to be called within qua.program() context manager!
        """
        qua.assign(
            self.state_gettable.qua_result_var,
            self.current_gettable.qua_result_var > self.arbok_params.threshold.qua
            )

    def _create_gettables(self):
        """Creates the gettables for the given readout"""
        self.current_gettable = self.get_gettable_from_path(self.charge_readout)

        self.state_gettable = self.create_gettable(
            gettable_name = self.name,
            var_type = bool,
        )
        self.gettables[self.state_gettable.full_name] = self.state_gettable




from arbok_driver.examples.sub_sequences import StabilityMap
from arbok_driver.examples.configurations import (
    stability_map_config
)
def test_stability_map(mock_measurement):
    stability_map = StabilityMap(
        mock_measurement, 'stability_map', stability_map_config
    )
    prg_str = mock_measurement.get_qua_program_as_str(recompile = True)
    assert len(stability_map.signals) == 1
    assert len(stability_map.readout_groups) == 1
    assert len(stability_map.abstract_readouts) == 1




# Single QUA script generated at 2026-03-28 03:11:46.787607
# QUA library version: 1.2.5


from qm import CompilerOptionArguments
from qm.qua import *

with program() as prog:
    v1 = declare(fixed, )
    v2 = declare(int, value=0)
    v3 = declare(int, )
    with infinite_loop_():
        pause()
        assign(v2, 0)
        assign(v3, 0)
        with while_((v3<10)):
            assign(v1, (0.0+Cast.mul_fixed_by_int(0.00999,v3)))
            align()
            align("gate_1", "gate_2", "gate_3", "gate_4")
            align("gate_1", "gate_2", "gate_3", "gate_4")
            play("unit_ramp"*amp(0.2), "gate_1", duration=20)
            play("unit_ramp"*amp(-0.1), "gate_2", duration=20)
            play("unit_ramp"*amp((v1-0)), "gate_3", duration=20)
            play("unit_ramp"*amp(0.25), "gate_4", duration=20)
            align("gate_1", "gate_2", "gate_3", "gate_4")
            wait(100, "gate_1", "gate_2", "gate_3", "gate_4")
            align("gate_1", "gate_2", "gate_3", "gate_4")
            play("unit_ramp"*amp(-0.2), "gate_1", duration=20)
            play("unit_ramp"*amp(0.1), "gate_2", duration=20)
            play("unit_ramp"*amp((0-v1)), "gate_3", duration=20)
            play("unit_ramp"*amp(-0.25), "gate_4", duration=20)
            align("gate_1", "gate_2", "gate_3", "gate_4")
            align()
            assign(v2, (v2+1))
            r0 = declare_stream()
            save(v2, r0)
            align()
            assign(v3, (v3+1))
    with stream_processing():
        r0.buffer(1).save("qm_driver2_mock_measurement2_shots")

config = None

loaded_config = None






import re

import pytest
from qm import generate_qua_script, qua

from arbok_driver import arbok

@pytest.mark.parametrize("nr_elements", [0, 1, 2, 4])
def test_correct_amount_of_resets(nr_elements):
    with qua.program() as prg:
        arbok.reset_sticky_elements([f"P{i}" for i in range(nr_elements)])
    prg_str = generate_qua_script(prg)
    assert len(list(re.finditer('ramp_to_zero', prg_str))) == nr_elements

def test_align():
    with qua.program() as prg:
        arbok.reset_sticky_elements(["P0", "P1"], do_align=True)
    prg_str = generate_qua_script(prg)
    assert len(list(re.finditer('align', prg_str))) == 1
    with qua.program() as prg:
        arbok.reset_sticky_elements(["P0", "P1"], do_align=False)
    prg_str = generate_qua_script(prg)
    assert len(list(re.finditer('align', prg_str))) == 0



"""
Module containing the ArbokDriver class for managing and running sequences
on a physical OPX instrument.
"""
from __future__ import annotations
from typing import TYPE_CHECKING
import copy

import xarray as xr
from qm import SimulationConfig, generate_qua_script
from qm.quantum_machines_manager import QuantumMachinesManager
from qcodes.instrument import Instrument
from qcodes.dataset import load_or_create_experiment
from sqlalchemy.orm import Session

from .device import Device
from .measurement import Measurement
from .sqlalchemy_classes import SqlRun

if TYPE_CHECKING:
    from .experiment import Experiment
    from .sqlalchemy_classes import SqlRun
    from qm.jobs.running_qm_job import RunningQmJob
    from qm.quantum_machine import QuantumMachine
    from qm.quantum_machines_manager import QuantumMachinesManager
    from qm import StreamsManager

class ArbokDriver(Instrument):
    """
    Class containing all functionality to manage and run modular sequences on a 
    physical OPX instrument
    """
    opx: QuantumMachine
    qm_job: RunningQmJob
    qmm: QuantumMachinesManager
    result_handles: StreamsManager

    def __init__(
            self,
            name: str,
            device: Device,
            **kwargs
            ) -> None:
        """
        Constructor class for `Program` class
        
        Args:
            name (str): Name of the instrument
            device (Device): Device class describing phyical device
            **kwargs: Arbitrary keyword arguments for qcodes Instrument class
        """
        super().__init__(name, **kwargs)
        if not isinstance(device, Device):
            raise TypeError(f"device must be type Device, is {type(Device)}")
        self.device: Device = device
        self.is_mock: bool = False
        self._measurements: list[Measurement] = []
        self.add_parameter('iteration', get_cmd = None, set_cmd =None)

        self.database_engine = None
        self.minio_filesystem = None
        self.station = None

    @property
    def measurements(self) -> list[Measurement]:
        """Measurements to be run within program uploaded to the OPX"""
        return self._measurements

    def reset_measurements(self) -> None:
        """
        Resets all measurements in the program
        TODO: delete instances of those measurements
        """
        for measurement in self._measurements:
            print(f"Deleting measurement: {measurement.short_name}")
            del measurement
        self._measurements = []
        self.submodules = {}

    def connect_opx(
            self,
            host_ip: str,
            qm_config: dict = None,
            reconnect: bool = False,
            **kwargs) -> None:
        """
        Creates QuantumMachinesManager and opens a quantum machine on it with
        the given IP address
        
        Args:
            host_ip (str): Ip address of the OPX
            qm_config (dict): QM/OPX config dictionary to be used. Defaults to
                None, in which case the config from the device is used. If given
                overwrites the device config.
            reconnect (bool): Whether to reconnect to the OPX if already
                connected. Keeps QMM alive if true. Defaults to False.
            **kwargs: Arbitrary keyword arguments for QMM instanciation
        """
        if not reconnect:
            self.qmm = QuantumMachinesManager(
                host = host_ip, **kwargs)
        if qm_config is not None:
            if not isinstance(qm_config, dict):
                raise ValueError(
                    "qm_config must be a dictionary, not a string")
            self.device.config = copy.deepcopy(qm_config)
        self.opx = self.qmm.open_qm(self.device.config)

    def reconnect_opx(
            self, host_ip: str, qm_config: dict = None) -> None:
        """
        Reconnects to the OPX with the given IP address and closes the previous
        connection
        
        Args:
            host_ip (str): Ip address of the OPX
            qm_config (dict): QM/OPX config dictionary to be used. Defaults to
                None, in which case the config from the device is used. If given
                overwrites the device config. 
        """
        if self.opx is not None:
            print('Closing previous connection')
            self.opx.close()
            self.qmm.close_all_quantum_machines()
        self.connect_opx(host_ip, qm_config, reconnect = True)

    def add_measurement(self, new_measurement: Measurement):
        """
        Adds a class which inherits `Measurement` to the program and adds it
        as a QCoDeS sub-module
        
        Args:
            new_measurement (Measurement): The instance which inherits
            Measurement to be added
        """
        self._measurements.append(new_measurement)

    def run(self, qua_program, **kwargs):
        """
        Sends the qua program for execution to the OPX and sets the programs 
        result handles 
        
        Args:
            qua_program (program): QUA program to be executed
        """
        self.qm_job = self.opx.execute(qua_program, **kwargs)
        self.result_handles = self.qm_job.result_handles

    def print_qua_program_to_file(
            self,
            path: str,
            qua_program,
            add_config: bool = False
            ) -> None:
        """
        Creates file with 'filename' and prints the QUA code to this file
        
        Args:
            file_name (str): File name of target file
            qua_program (program): QUA program to be printed
            add_config (bool): Whether config is added to output file
        """
        with open(path, 'w', encoding="utf-8") as file:
            if self.device is not None and add_config:
                file.write(generate_qua_script(
                    qua_program, self.device.config
                    ))
            else:
                file.write(generate_qua_script(qua_program))

    def run_local_simulation(
            self,
            qua_program,
            duration: int,
            **kwargs
            ):
        """
        Simulates the given program of the sequence for `duration` cycles
        TODO: Move to SequenceBase and add checks if OPX is connected
        Args:
            qua_program (program): QUA program to be simulated
            duration (int): Simulation duration in cycles
            **kwargs: Arbitrary keyword arguments for QMM simulation

        Returns:
            simulated_job (SimulatedJob): QM job with waveform simulation result
            nr_controllers (int): Nr of controllers to fetch simulation results
        """
        if not self.qmm:
            raise ConnectionError(
                "No QMM found! Connect an OPX via `connect_OPX`")
        simulated_job = self.qmm.simulate(
            self.opx.get_config(),
            qua_program,
            SimulationConfig(duration=duration),
            **kwargs
        )
        return simulated_job

    def get_idn(self):
        """
        Overload the get_idn method as we don't have one.
        """
        return None

    def ask_raw(self, cmd: str) -> str:
        """Abstract method from qcodes Instrument"""
        raise NotImplementedError

    def write_raw(self, cmd: str) -> None:
        """Abstract method from qcodes Instrument"""
        raise NotImplementedError

    def create_measurement_from_experiment(
            self,
            experiment: Experiment,
            qc_measurement_name: str | None = None,
            name: str = 'measurement',
            ) -> Measurement:
        """
        Creates an arbok and QCoDeS measurement from an arbok-experiment
        
        Args:
            name (str): Name of the measurement (py. variable name compliant)
            experiment (arbok_driver.Experiment): Experiment to be run
            qc_measurement_name (str): Name of the QCoDeS measurement
                (as it will be saved in the database)

        Returns:
            measurement (arbok_driver.Measurement): Measurement instance
        """
        measurement = Measurement(
            parent = self,
            name = name,
            sequence_config = self.device.param_config
            )
        if qc_measurement_name is None:
            qc_measurement_name = experiment.name
        measurement.qc_experiment = load_or_create_experiment(
            experiment.name, self.device.name)
        measurement.qc_measurement_name = qc_measurement_name
        sub_sequences = experiment.get_sequences_config(self.device)
        measurement.add_subsequences_from_dict(sub_sequences)
        return measurement

    def check_db_engine_and_bucket_connected(self):
        """
        Checks if database engine and s3 bucket are connected
        Raises error if not connected
        TODO: ping both TCP connections to check if still alive
        """
        if self.database_engine is None:
            raise ConnectionError(
                "No database engine connected! Please connect a database "
                "engine to the arbok_driver before running measurements.")
        if self.minio_filesystem is None:
            raise ConnectionError(
                "No MinIO filesystem connected! Please connect a MinIO filesystem to the "
                "arbok_driver before running measurements.")

    def get_run_from_id(self, run_id: int) -> SqlRun:
        """
        Fetches a run from the connected arbok database engine based on the
        given run ID
        
        Args:
            run_id (int): ID of the run to be fetched
        Returns:
            run (arbok_driver.sqlalchemy_classes.SqlRun): detached run instance
        """
        if self.database_engine is None:
            raise ConnectionError(
                "No database engine connected! Please connect a database "
                "engine to the arbok_driver before fetching runs.")
        with Session(self.database_engine) as session:
            sql_run = session.get(SqlRun, run_id)
        if sql_run is None:
            raise LookupError(
                f"No SqlRun found for run-ID: {run_id}"
            )
        return sql_run

    def get_data_from_id(self, run_id: int) -> xr.Dataset:
        """
        Fetches data from the connected database engine based on the
        given run ID
        
        Args:
            run_id (int): ID of the run to fetch data from
        Returns:
            xr_dataset (xarray.Dataset): Lazy loaded xarray dataset. Will only
                load data when using .load() or .compute() methods!
        """
        sql_run = self.get_run_from_id(run_id)
        minio_name = f"{sql_run.run_id}_{sql_run.uuid}"
        store = self.minio_filesystem.get_mapper(f'dev/{minio_name}/data.zarr')
        xr_dataset = xr.open_zarr(store, consolidated = True)
        return xr_dataset



'''Module containing ReadSequence implementing coulomb peaks measurement'''
from dataclasses import dataclass
from qm import qua
from qm.qua._expressions import QuaVariable
from arbok_driver import (
    arbok, ParameterClass, ReadSequence, SequenceBase
)
from arbok_driver.parameter_types import (
    Time, ParameterMap, List, Voltage
)

@dataclass(frozen=True)
class CoulombPeaksParameters(ParameterClass):
    t_wait_after_ramp: Time
    t_wait_before_chop: Time
    gate_elements: List
    set_elements: List
    v_home: ParameterMap[str, Voltage]
    v_set_level: ParameterMap[str, Voltage]

class CoulombPeaks(ReadSequence):
    """
    Class containing parameters and sequence for Coulomb peaks measurement of
    a generic SET
    """
    PARAMETER_CLASS = CoulombPeaksParameters
    arbok_params: CoulombPeaksParameters
    qua_rep_index: QuaVariable

    def __init__(
            self,
            parent: SequenceBase,
            name: str,
            sequence_config: dict,
            reset_gates_only: bool = False
    ):
        """
        Constructor method for 'MixedDownUpInit' class

        Args:
            parent (SequenceBase): Parent sequence to be added to
            name (str): name of sequence
            sequence_config (dict): config containing pulse parameters
            reset_gates_only (bool): True to reset the gates only, not present
                or false to reset SET too
        """
        super().__init__(
            parent = parent,
            name = name,
            sequence_config = sequence_config
        )
        self.g_elements = list(self.arbok_params.gate_elements.get())
        self.elements = self.g_elements + list(self.arbok_params.set_elements.get())
        self.gates_only = reset_gates_only

    def qua_declare(self):
        super().qua_declare()
        self.qua_rep_index = qua.declare(int)

    def qua_sequence(self):
        qua.align(*self.elements)

        ### Go from the set 'home' voltage point to the 'level' voltage point
        arbok.ramp(
            elements = self.elements,
            reference = self.arbok_params.v_home,
            target = self.arbok_params.v_set_level,
            operation = 'unit_ramp',
        )
        with qua.for_(
            var = self.qua_rep_index,
            init = 0,
            cond = self.qua_rep_index < 4,
            update = self.qua_rep_index + 1
        ):
            qua.wait(self.arbok_params.t_wait_after_ramp.qua, *self.elements)
        qua.align(*self.elements)

        ### Measure the SET current
        for _, readout in self.readout_groups["read"].items():
            readout.qua_measure_and_save()

        qua.align(*self.elements)
        qua.wait(self.arbok_params.t_wait_before_chop.qua, *self.elements)

        ### Chopped readout with one iteration to get the peak derivative
        for _, readout in self.readout_groups["chop"].items():
            readout.qua_measure_and_save()
        qua.align(*self.elements)

        ### Go back to the set 'level' voltage point to the 'home' voltage point
        arbok.ramp(
            elements = self.elements,
            reference = self.arbok_params.v_set_level,
            target = self.arbok_params.v_home,
            operation = 'unit_ramp',
        )
        if self.gates_only:
            arbok.reset_sticky_elements(self.g_elements)
        else:
            arbok.reset_sticky_elements(self.elements)



"""Module containing test for MeasurementRunner class"""
import pytest

import numpy as np

def test_qcodes_measurement_runner(stability_map, mock_measurement) -> None:
    mock_measurement.driver.is_mock = True
    mock_measurement.set_sweeps(
        {stability_map.v_home_P1: np.arange(10)},
        {stability_map.v_home_P2: np.arange(5)}
        )
    mock_measurement.register_gettables(keywords = ['_'])
    nr_gettables = len(mock_measurement.gettables)
    mock_measurement.qc_measurement_name = 'test_measurement'
    mock_measurement.run_measurement(
        ext_sweep_list=[{stability_map.v_home_P3: np.arange(2)}])
    dataset = mock_measurement.dataset
    assert len(dataset) == nr_gettables
    first_dataarray = next(iter(dataset.data_vars.values()))
    assert first_dataarray.shape == (2, 10, 5)



from . import path_finders
from . import utils
from .abstract_readout import AbstractReadout
from .arbok_driver import ArbokDriver
from .device import Device
from .ekans import Ekans
from .experiment import Experiment
from .generic_tunig_interface import GenericTuningInterface
from .parameters.gettable_parameter import GettableParameter
from .parameter_class import EmptyParameterClass, ParameterClass
from .measurement import Measurement
from .read_sequence import ReadSequence
from .sequence_base import SequenceBase
from .parameters.sequence_parameter import SequenceParameter
from .signal import Signal
from .sub_sequence import SubSequence
from .sweep import Sweep



"""Module containing signal class"""
from __future__ import annotations
from typing import TYPE_CHECKING
import logging

if TYPE_CHECKING:
    from arbok_driver import ReadSequence
    from arbok_driver.parameters import GettableParameterBase

class Signal:
    """
    Class managing voltage signal from certain readout elements of OPX
    """
    def __init__(
        self,
        name: str,
        read_sequence: ReadSequence,
        ):
        """
        Constructor method of Signal class
        a
        Args:
            name (str): Name of the signal
            read_sequence (arbok_driver.ReadSequence): ReadSequence performing the
                abstract readout
        """
        self.name: str = name
        self.read_sequence: ReadSequence = read_sequence
        self._gettables: dict[str, GettableParameterBase] = {}

    @property
    def gettables(self) -> dict[str, GettableParameterBase]:
        """Dictionary with all gettables registered on the signal"""
        return self._gettables

    def add_gettable(self, gettable: GettableParameterBase) -> None:
        """
        Adds a gettable parameter to the signal

        Args:
            gettable (GettableParameter): Gettable parameter to be added
        """
        logging.debug(
            "Adding gettable %s to signal %s", gettable.name, self.name)
        if gettable.name in self._gettables:
            raise ValueError(
                f"gettable with name {gettable.name} already exists in"
                f" signal '{self.name}'."
            )
        self._gettables[gettable.name] = gettable
        if hasattr(self, gettable.name):
            raise ValueError(
                f"Attribute with name {gettable.name} already exists on"
                f" signal '{self.name}'. Choose a different name for the gettable."
            )
        setattr(self, gettable.name, gettable)



""" Module containing Sequence class """
from __future__ import annotations
from abc import ABC
from dataclasses import fields
from typing import TYPE_CHECKING, Any
import logging

from .parameter_class import ParameterClass
from .sequence_base import SequenceBase

if TYPE_CHECKING:
    from .measurement import Measurement

class SubSequence(SequenceBase, ABC):
    """
    Class describing a subsequence of a QUA programm (e.g Init, Control, Read). 
    """
    _enforce_parameter_class: bool = False
    def __init__(
            self,
            parent,
            name: str,
            sequence_config: dict | None = None,
            check_step_requirements: bool = False,
            **kwargs
            ):
        """
        Constructor class for `Program` class
        
        Args:
            name (str): Name of the program
            sequence_config (dict): Dictionary containing all device parameters
            check_step_requirements (bool): Whether to check step requirements
                for this subsequence
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(
            parent, name, sequence_config, check_step_requirements, **kwargs)
        self.parent.add_subsequence(self)
        self.arbok_params = self.map_arbok_params()

    @property
    def measurement(self) -> Measurement:
        """Returns parent (sub) sequence"""
        return self.find_measurement()

    def qua_sequence(self):
        if self.check_step_requirements:
            self.measurement.qua_check_step_requirements(
                super().qua_sequence
            )
        else:
            super().qua_sequence()

    def map_arbok_params(self) -> ParameterClass:
        """Maps required params for QUA code to arbok_params attribute"""
        arg_names = [f.name for f in fields(self.PARAMETER_CLASS) if f.init]
        init_dict = self.measurement.get_parameters_and_maps(arg_names)
        init_dict.update(self.get_parameters_and_maps(arg_names))
        return self.PARAMETER_CLASS(**init_dict)

    def add_subsequences_from_dict(
            self,
            subsequence_dict: dict,
            namespace_to_add_to: dict | None = None) -> None:
        """
        Adds subsequences to the sequence from a given dictionary

        Args:
            subsequence_dict (dict): Dictionary containing the subsequences
            namespace_to_add_to (dict): Namespace to add the registered
                subsequences to
        """
        class ContainerParameterClass(ParameterClass):
            pass

        class ContainerSubSequence(SubSequence):
            PARAMETER_CLASS = ContainerParameterClass
    
        super()._add_subsequences_from_dict(
            default_sequence = ContainerSubSequence,
            subsequence_dict = subsequence_dict,
            namespace_to_add_to = namespace_to_add_to
        )

    def find_measurement(self) -> Measurement:
        """Recursively searches the parent sequence"""
        if self.parent.__class__.__name__ == 'Measurement':
            return self.parent
        elif isinstance(self.parent, SubSequence):
            return self.parent.find_measurement()
        else:
            raise ValueError(
                "Parent sequence must be of type Sequence. "
                f"Is of type {self.parent.__class__.__name__}")

    def get_sequence_path(self, path: str | None = None) -> str:
        """Returns the path of subsequences up to the parent sequence"""
        if path is None:
            path = ""
        if self.parent.__class__.__name__ == 'Measurement':
            return f"{self.parent.short_name}__{self.short_name}__{path}"
        elif self.parent is None:
            return f"{self.short_name}__{path}"
        else:
            return self.parent.get_sequence_path(f"{self.short_name}__{path}")

    def _return_measurement_parameters(self, key: str) -> Any:
        """Returns attribute from parent sequence"""
        logging.debug("Searching parent sequence %s for parameter %s",
                        self.measurement.name, key)
        if key in self.measurement.parameters:
            return self.measurement.parameters[key]
        raise AttributeError(
            f"Parameter {key} not found in {self.measurement.name}")



from arbok_driver.parameter_types import Voltage, Time, String
from arbok_driver.examples.sub_sequences import SquarePulse

square_pulse_config = {
    'sequence': SquarePulse,
    'parameters': {
        'amplitude': {
            'type': Voltage,
            'value': 0.1,
        },
        't_square_pulse': {
            'type': Time,
            'value': 100,
        },
        't_ramp': {
            'type': Time,
            'value': 200,
        },
        'element': {
            'type': String,
            'value': 'gate_1',
        }
    }
}



"""Test module asserting functionality of arbok.ramp"""
from dataclasses import dataclass
import re

import pytest
from qm import generate_qua_script, qua

from arbok_driver import arbok, ParameterClass, SubSequence
from arbok_driver.parameter_types import Voltage, ParameterMap

@dataclass(frozen = True)
class TestSubSequenceParameterClass(ParameterClass):
    v_home0: ParameterMap[str, Voltage]
    v_home1: ParameterMap[str, Voltage]
    v_home2: ParameterMap[str, Voltage]
    v_home3: ParameterMap[str, Voltage]

class UserSubSequence(SubSequence):
    PARAMETER_CLASS = TestSubSequenceParameterClass
    arbok_params: TestSubSequenceParameterClass
    
test_conf = {
    "sequence": UserSubSequence,
    "parameters": {
        'v_home0': {
            'type': Voltage,
            'elements': {
                'P1': 0.0,
                'P2': 0.0,
                'P3': 0.0
            }
        },
        'v_home1': {
            'type': Voltage,
            'elements': {
                'P1': 0.0,
                'P2': 0.0,
            }
        },
        'v_home2': {
            'type': Voltage,
            'elements': {
                'P1': 0.0,
                'P2': 0.0,
                'P3': 0.0
            }
        },
        'v_home3': {
            'type': Voltage,
            'elements': {
                'P1': 0.1,
                'P2': 0.2,
                'P3': 0.3
            }
        },
    }
}

def test_nr_of_ramps(square_pulse_scalable, mock_measurement):
    prog_str = mock_measurement.get_qua_program_as_str(recompile = True)
    assert len(list(re.finditer("unit_ramp", prog_str))) == 6

    square_pulse_scalable.arbok_params.sticky_elements.set(["P1", "P2"])
    prog_str = mock_measurement.get_qua_program_as_str(recompile = True)
    assert len(list(re.finditer("unit_ramp", prog_str))) == 4

def test_no_ramps_for_negligible_amplitudes(
        square_pulse_scalable, mock_measurement):
    square_pulse_scalable.arbok_params.v_square["P1"].set(0)
    square_pulse_scalable.arbok_params.v_square["P2"].set(0)
    prog_str = mock_measurement.get_qua_program_as_str(recompile = True)
    assert len(list(re.finditer("unit_ramp", prog_str))) == 2

def test_ramp_without_origin(mock_measurement):
    test_sequence = UserSubSequence(
        mock_measurement, "test_sequence", test_conf
        )
    with qua.program() as prg:
        arbok.ramp(
            elements = ["P1", "P2"],
            target = test_sequence.arbok_params.v_home3,
            operation = "unit_ramp"
        )
    prg_str = generate_qua_script(prg)
    print(prg_str)
    assert len(list(re.finditer("unit_ramp", prg_str))) == 2

def test_missing_elements_in_mapping(mock_measurement):
    test_sequence = UserSubSequence(
        mock_measurement, "test_sequence", test_conf
        )
    with pytest.raises(ValueError):
        arbok.ramp(
            elements = ["P1", "P2", "P12"],
            target = test_sequence.arbok_params.v_home0,
            operation = "unit_ramp"
        )

def test_difference_calulated_correctly(mock_measurement):
    test_sequence = UserSubSequence(
        mock_measurement, "test_sequence", test_conf
        )
    with qua.program() as prg:
        arbok.ramp(
            elements = ["P1"],
            target = test_sequence.arbok_params.v_home3,
            reference = test_sequence.arbok_params.v_home0,
            operation = "unit_ramp"
        )
    prg_str = generate_qua_script(prg)
    amp = prg_str.split("amp(")[1].split(')')[0]
    assert float(amp) == pytest.approx(0.6)



"""
Module containing configuration of coulomb peak measurement on set1 and set2
"""

from arbok_driver.parameter_types import Voltage, Time, Int, List
from arbok_driver.examples.readout_classes import (
    DcAverage,
    DcChoppedReadout,
)
from arbok_driver.examples.sub_sequences import CoulombPeaks

coulomb_peaks_set1_set2_conf = {
    'sequence': CoulombPeaks,
    'parameters': {
        'set_elements': {
            'type': List,
            'value': ['SET1', 'SET2']
        },
        'gate_elements': {
            'type': List,
            'value': [
                'P1', 'J1',
                'P2', 'J2',
                'P3', 'J3',
                'P4', 'J4',
                'P5', 'J5',
                'P6', 'J6',
                'P7', 'J7',
                'P8',
            ]
        },
        't_wait_after_ramp': {
            'type': Time,
            'value': int(1e6/4)
        },
        't_wait_before_chop': {
            'type': Time,
            'value': int(100e3/4)
        },
        'v_home': {
            'type': Voltage,
            'elements': {
                'SET1': 0, 'SET2': 0,
                'P1': 0, 'J1': 0,
                'P2': 0, 'J2': 0,
                'P3': 0, 'J3': 0,
                'P4': 0, 'J4': 0,
                'P5': 0, 'J5': 0,
                'P6': 0, 'J6': 0,
                'P7': 0, 'J7': 0,
                'P8': 0,
            }
        },
        'v_set_level': {
            'type': Voltage,
            'elements': {
                'SET1': 0, 'SET2': 0,
                'P1': 0.1, 'J1': 0,
                'P2': 0, 'J2': 0,
                'P3': 0, 'J3': 0,
                'P4': 0, 'J4': 0,
                'P5': 0, 'J5': 0,
                'P6': 0, 'J6': 0,
                'P7': 0, 'J7': 0,
                'P8': 0,
            }
        },
    },
    'signals': ['set1', 'set2', 'chopped_readouts'],
    'readout_groups': {
        'read': {
            'set1': {
                'readout_class': DcAverage,
                'signal': 'set1',
                'kwargs': {
                    'qua_element': 'SET1'
                }
            },
            'set2': {
                'readout_class': DcAverage,
                'signal': 'set2',
                'kwargs': {
                    'qua_element': 'SET2'
                }
            }
        },
        'chop': {
            'stability': {
                'readout_class': DcChoppedReadout,
                'signal': 'chopped_readouts',
                'kwargs': {
                    'readout_qua_elements': {
                        'p1p2': 'SET1',
                        'p7p8': 'SET2',
                    }
                },
                'parameters': {
                    'v_chop': {
                        'type': Voltage,
                        'elements': {
                            'P1': 0.01,
                            'P8': 0.01,
                        }
                    },
                    'n_chops': {
                        'type': Int,
                        'value': 20
                    },
                    'chop_wait': {
                        'type': Time,
                        'value': int(20e3/4)
                    },
                },
            }
        },
    }
}



import importlib
import sys
import types
import pytest

from arbok_driver.ekans import Ekans  # adjust import if needed


# Use your real examples package
import arbok_driver.examples as examples


# ----------------------------
# Basic initialization
# ----------------------------

def test_init_with_valid_module():
    ek = Ekans(examples)
    assert ek is not None


def test_init_with_invalid_type():
    with pytest.raises(TypeError):
        Ekans("not_a_module")


# ----------------------------
# Module collection
# ----------------------------

def test_collects_modules():
    ek = Ekans(examples)
    modules = ek._collect_module_names()

    # sanity checks
    assert any("examples.configurations" in m for m in modules)
    assert any("examples.sub_sequences" in m for m in modules)
    assert any("examples.readout_classes" in m for m in modules)


# ----------------------------
# Attribute tree structure
# ----------------------------

def test_namespace_structure_exists():
    ek = Ekans(examples)

    assert hasattr(ek, "configurations")
    assert hasattr(ek, "experiments")
    assert hasattr(ek, "readout_classes")
    assert hasattr(ek, "sub_sequences")


def test_nested_namespace():
    ek = Ekans(examples)

    assert hasattr(ek.configurations, "hardware_configs")
    assert hasattr(ek.configurations, "sequence_configs")


# ----------------------------
# Flattening behavior
# ----------------------------

def test_flattened_attributes_from_submodules():
    ek = Ekans(examples)

    # e.g. square_pulse.py should expose things directly under sub_sequences
    attrs = dir(ek.sub_sequences)

    # We don't know exact class/function names, so just ensure flattening happened
    assert len(attrs) > 0
    assert not any(name.endswith(".py") for name in attrs)


def test_no_private_attributes_exposed():
    ek = Ekans(examples)

    attrs = dir(ek.sub_sequences)

    # ignore dunder attributes, only check "real" ones
    public_like = [name for name in attrs if not (name.startswith("__") and name.endswith("__"))]

    assert not any(name.startswith("_") for name in public_like)


# ----------------------------
# __all__ behavior
# ----------------------------

def test_respects___all__():
    ek = Ekans(examples)

    module = importlib.import_module(
        "arbok_driver.examples.readout_classes.dc_average"
    )

    if hasattr(module, "__all__"):
        for name in module.__all__:
            assert hasattr(ek.readout_classes, name)


# ----------------------------
# Reload behavior
# ----------------------------

def test_reload_modules_runs_without_error():
    ek = Ekans(examples)

    ek.reload_modules()  # should not raise


def test_reload_preserves_structure():
    ek = Ekans(examples)

    before = dir(ek.sub_sequences)
    ek.reload_modules()
    after = dir(ek.sub_sequences)

    assert before == after


# ----------------------------
# Submodule attachment
# ----------------------------

def test_direct_submodule_attributes_are_attached():
    ek = Ekans(examples)

    module = importlib.import_module(
        "arbok_driver.examples.sub_sequences.square_pulse"
    )

    public = getattr(module, "__all__", None)
    if public is None:
        public = [n for n in dir(module) if not n.startswith("_")]

    found = [name for name in public if hasattr(ek.sub_sequences, name)]

    # At least one should be attached
    assert len(found) > 0


# ----------------------------
# No overwrite behavior
# ----------------------------

def test_no_overwrite_of_existing_attributes():
    ek = Ekans(examples)

    attrs = dir(ek.sub_sequences)

    # ensure uniqueness
    assert len(attrs) == len(set(attrs))


# ----------------------------
# Internal helper correctness
# ----------------------------

def test_get_submodules_only_direct_children():
    ek = Ekans(examples)

    module = importlib.import_module(
        "arbok_driver.examples.sub_sequences"
    )

    subs = ek._get_submodules(module)

    for sub in subs:
        assert sub.__name__.count(".") == module.__name__.count(".") + 1


# ----------------------------
# Robustness: broken module import
# ----------------------------

def test_import_error_propagates(monkeypatch):
    ek = Ekans(examples)

    # remove one module so import_module gets called
    target = "arbok_driver.examples.sub_sequences.square_pulse"
    sys.modules.pop(target, None)

    def broken_import(name):
        raise ImportError("boom")

    monkeypatch.setattr(importlib, "import_module", broken_import)

    with pytest.raises(ImportError):
        ek.reload_modules()


"""Module containing ReadSequence implementing PSB readout"""
from dataclasses import dataclass

from qm import qua
from arbok_driver import (
    arbok, ParameterClass, SequenceBase, ReadSequence
)
from arbok_driver.parameter_types import (
    Time, ParameterMap,  Voltage, List
)

@dataclass(frozen = True)
class ParityReadParameters(ParameterClass):
    v_home: ParameterMap[str, Voltage]
    v_read: ParameterMap[str, Voltage]
    v_reference: ParameterMap[str, Voltage]
    gate_elements: List
    readout_elements: List
    t_wait_home_before: Time
    t_wait_pre_read: Time
    t_ramp_to_reference: Time
    t_ramp_to_read: Time
    t_wait_post_read: Time
    t_wait_after_reset: Time

class ParityRead(ReadSequence):
    """
    Class containing parameters and sequence for spin parity readout
    Args:
        param_config (dict): Dict containing all program parameters 
    """
    PARAMETER_CLASS= ParityReadParameters
    arbok_params: ParityReadParameters

    def __init__(
            self,
            parent: SequenceBase,
            name: str,
            sequence_config: dict,
    ):
        """
        Constructor method for 'ParityRead' class
        
        Args:
            parent (SequenceBase): Parent instrument module
                or instrument
            name (str): name of the sequence
            sequence_config (dict): Configuration for sequence
        """
        super().__init__(
            parent = parent,
            name = name,
            sequence_config = sequence_config,
            )
        self.elements = list(self.arbok_params.gate_elements.get())
        self.elements += list(self.arbok_params.readout_elements.get())

    def qua_declare(self):
        """
        Declares variables before execution of the main program. All mandatory
        qua variables for auto generated readouts from the config are introduced
        by the parent class (super)
        """
        return super().qua_declare()

    def qua_sequence(self):
        """
        QUA sequence to perform spin parity readout
        
        The sequence is as follows:
            1. Move to REFERENCE measurement point
            2. Take physical REFERENCE measurement
            3. Move to READ measurement point
            4. Take physical READ measurement
        """
        qua.align()
        qua.align(*self.elements)
        qua.wait(self.arbok_params.t_wait_home_before.qua, *self.elements)
        ### Move to REFERENCE measurement point
        arbok.ramp(
            elements = self.arbok_params.gate_elements.get(),
            reference = self.arbok_params.v_home,
            target = self.arbok_params.v_reference,
            duration = self.t_ramp_to_reference,
            operation = 'unit_ramp',
            )
        qua.wait(self.arbok_params.t_wait_pre_read.qua, *self.elements)
        qua.align(*self.elements)
        ### Take physical REFERENCE measurement
        for _, readout in self.readout_groups["ref"].items():
            readout.qua_measure()
        qua.align(*self.elements)
        qua.wait(self.t_wait_post_read.qua, *self.elements)

        ### Move to READ measurement point
        arbok.ramp(
            elements = self.arbok_params.gate_elements.get(),
            reference = self.arbok_params.v_reference,
            target = self.arbok_params.v_read,
            duration = self.arbok_params.t_ramp_to_read,
            operation = 'unit_ramp',
            )
        qua.wait(self.arbok_params.t_wait_pre_read.qua, *self.elements)
        qua.align(*self.elements)
        ### take physical READ measurement
        for _, readout in self.readout_groups["read"].items():
            readout.qua_measure()
        qua.align(*self.elements)
        qua.align()

        ### Process raw data in abstract readouts
        for _, readout in self.readout_groups["diff"].items():
            readout.qua_measure()
        for _, readout in self.readout_groups["state"].items():
            readout.qua_measure()
        qua.align(*self.elements)
        qua.wait(self.arbok_params.t_wait_post_read.qua, *self.elements)

        ### Reset elements
        arbok.reset_sticky_elements(self.arbok_params.gate_elements.get())
        qua.wait(self.arbok_params.t_wait_after_reset.qua, *self.elements)
        qua.align(*self.elements)

        ### Apply feedback if given in configuration
        if 'set_feedback' in self.readout_groups:
            for _, readout in self.readout_groups["set_feedback"].items():
                readout.qua_measure()
        qua.align()

    def qua_after_sequence(self):
        """Saves variables to the respective streams after the sequence"""
        self.measurement.qua_check_step_requirements(self.save_variables)

    def save_variables(self):
        """Saves all variables to the respective streams"""
        for _, readout in self.abstract_readouts.items():
            readout.qua_save_variables()




# Single QUA script generated at 2026-03-28 05:10:54.432098
# QUA library version: 1.2.5


from qm import CompilerOptionArguments
from qm.qua import *

with program() as prog:
    v1 = declare(int, )
    v2 = declare(int, value=0)
    v3 = declare(int, )
    input_stream_mock_measurement_int_input_stream = declare_input_stream(int, 'mock_measurement_int_input_stream', size=1)
    v4 = declare(fixed, )
    input_stream_mock_measurement_fixed_input_stream = declare_input_stream(fixed, 'mock_measurement_fixed_input_stream', size=1)
    v5 = declare(fixed, )
    v6 = declare(fixed, )
    v7 = declare(fixed, )
    v8 = declare(fixed, )
    v9 = declare(fixed, )
    v10 = declare(fixed, )
    v11 = declare(bool, )
    v12 = declare(bool, )
    v13 = declare(fixed, )
    v14 = declare(fixed, )
    v15 = declare(fixed, )
    v16 = declare(fixed, )
    v17 = declare(fixed, )
    v18 = declare(fixed, )
    v19 = declare(bool, )
    v20 = declare(bool, )
    v21 = declare(int, )
    with infinite_loop_():
        pause()
        assign(v2, 0)
        advance_input_stream(input_stream_mock_measurement_int_input_stream)
        assign(v3, input_stream_mock_measurement_int_input_stream[0])
        advance_input_stream(input_stream_mock_measurement_fixed_input_stream)
        assign(v4, input_stream_mock_measurement_fixed_input_stream[0])
        assign(v21, 0)
        with while_((v21<100)):
            assign(v1, (0+(1*v21)))
            align()
            align("P1", "J1", "P2", "P7", "J7", "P8")
            play("unit_ramp"*amp(0.06), "P1", duration=250)
            play("unit_ramp"*amp(-0.085), "J1", duration=250)
            play("unit_ramp"*amp(-0.06), "P2", duration=250)
            play("unit_ramp"*amp(-0.07), "P7", duration=250)
            play("unit_ramp"*amp(-0.05), "J7", duration=250)
            play("unit_ramp"*amp(0.07), "P8", duration=250)
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(50000, "P1", "J1", "P2", "P7", "J7", "P8")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            play("unit_ramp"*amp(-0.0165), "P1", duration=250)
            play("unit_ramp"*amp(0.035), "J1", duration=250)
            play("unit_ramp"*amp(0.0165), "P2", duration=250)
            play("unit_ramp"*amp(0.099), "P7", duration=250)
            play("unit_ramp"*amp(0.117), "J7", duration=250)
            play("unit_ramp"*amp(-0.099), "P8", duration=250)
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(9571, "P1", "J1", "P2", "P7", "J7", "P8")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            play("unit_ramp"*amp(-0.03), "P1", duration=141)
            play("unit_ramp"*amp(0.03), "P2", duration=141)
            play("unit_ramp"*amp(-0.02), "P7", duration=141)
            play("unit_ramp"*amp(0.02), "P8", duration=141)
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(250, "P1", "J1", "P2", "P7", "J7", "P8")
            ramp_to_zero("P1")
            ramp_to_zero("J1")
            ramp_to_zero("P2")
            ramp_to_zero("P7")
            ramp_to_zero("J7")
            ramp_to_zero("P8")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            align()
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            wait(25000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            play("unit_ramp"*amp(0.023), "P1", duration=6250)
            play("unit_ramp"*amp(-0.02), "J1", duration=6250)
            play("unit_ramp"*amp(-0.023), "P2", duration=6250)
            play("unit_ramp"*amp(-0.0325), "P7", duration=6250)
            play("unit_ramp"*amp(0.005), "J7", duration=6250)
            play("unit_ramp"*amp(0.0325), "P8", duration=6250)
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(25000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            measure("measure", "SET1", integration.full("x_const", v5, ""))
            measure("measure", "SET2", integration.full("x_const", v6, ""))
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            wait(5000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            play("unit_ramp"*amp((v4-0.023)), "P1", duration=250)
            play("unit_ramp"*amp(((v4*-1)--0.023)), "P2", duration=250)
            play("unit_ramp"*amp(-0.0025000000000000022), "P7", duration=250)
            play("unit_ramp"*amp(0.0025000000000000022), "P8", duration=250)
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(25000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            measure("measure", "SET1", integration.full("x_const", v7, ""))
            measure("measure", "SET2", integration.full("x_const", v8, ""))
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align()
            assign(v9, (v5-v7))
            assign(v10, (v6-v8))
            assign(v11, (v9>0.001))
            assign(v12, (v10>-0.001))
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            wait(5000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            ramp_to_zero("P1")
            ramp_to_zero("J1")
            ramp_to_zero("P2")
            ramp_to_zero("P7")
            ramp_to_zero("J7")
            ramp_to_zero("P8")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(1000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align()
            align("P1", "J1", "P2", "P7", "J7", "P8")
            play("unit_ramp"*amp(0.06), "P1", duration=250)
            play("unit_ramp"*amp(-0.085), "J1", duration=250)
            play("unit_ramp"*amp(-0.06), "P2", duration=250)
            play("unit_ramp"*amp(-0.07), "P7", duration=250)
            play("unit_ramp"*amp(-0.05), "J7", duration=250)
            play("unit_ramp"*amp(0.07), "P8", duration=250)
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(50000, "P1", "J1", "P2", "P7", "J7", "P8")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            play("unit_ramp"*amp(-0.0165), "P1", duration=250)
            play("unit_ramp"*amp(0.035), "J1", duration=250)
            play("unit_ramp"*amp(0.0165), "P2", duration=250)
            play("unit_ramp"*amp(0.099), "P7", duration=250)
            play("unit_ramp"*amp(0.117), "J7", duration=250)
            play("unit_ramp"*amp(-0.099), "P8", duration=250)
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(9571, "P1", "J1", "P2", "P7", "J7", "P8")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            play("unit_ramp"*amp(-0.03), "P1", duration=v3)
            play("unit_ramp"*amp(0.03), "P2", duration=v3)
            play("unit_ramp"*amp(-0.02), "P7", duration=v3)
            play("unit_ramp"*amp(0.02), "P8", duration=v3)
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(250, "P1", "J1", "P2", "P7", "J7", "P8")
            ramp_to_zero("P1")
            ramp_to_zero("J1")
            ramp_to_zero("P2")
            ramp_to_zero("P7")
            ramp_to_zero("J7")
            ramp_to_zero("P8")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            align()
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            wait(25000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            play("unit_ramp"*amp(0.023), "P1", duration=6250)
            play("unit_ramp"*amp(-0.02), "J1", duration=6250)
            play("unit_ramp"*amp(-0.023), "P2", duration=6250)
            play("unit_ramp"*amp(-0.0325), "P7", duration=6250)
            play("unit_ramp"*amp(0.005), "J7", duration=6250)
            play("unit_ramp"*amp(0.0325), "P8", duration=6250)
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(25000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            measure("measure", "SET1", integration.full("x_const", v13, ""))
            measure("measure", "SET2", integration.full("x_const", v14, ""))
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            wait(5000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            play("unit_ramp"*amp((v4-0.023)), "P1", duration=250)
            play("unit_ramp"*amp(((v4*-1)--0.023)), "P2", duration=250)
            play("unit_ramp"*amp(-0.0025000000000000022), "P7", duration=250)
            play("unit_ramp"*amp(0.0025000000000000022), "P8", duration=250)
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(25000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            measure("measure", "SET1", integration.full("x_const", v15, ""))
            measure("measure", "SET2", integration.full("x_const", v16, ""))
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align()
            assign(v17, (v13-v15))
            assign(v18, (v14-v16))
            assign(v19, (v17>0.001))
            assign(v20, (v18>-0.001))
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            wait(5000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            ramp_to_zero("P1")
            ramp_to_zero("J1")
            ramp_to_zero("P2")
            ramp_to_zero("P7")
            ramp_to_zero("J7")
            ramp_to_zero("P8")
            align("P1", "J1", "P2", "P7", "J7", "P8")
            wait(1000, "P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align("P1", "J1", "P2", "P7", "J7", "P8", "SET1", "SET2")
            align()
            r2 = declare_stream()
            save(v5, r2)
            r3 = declare_stream()
            save(v6, r3)
            r4 = declare_stream()
            save(v7, r4)
            r5 = declare_stream()
            save(v8, r5)
            r6 = declare_stream()
            save(v9, r6)
            r7 = declare_stream()
            save(v10, r7)
            r8 = declare_stream()
            save(v11, r8)
            r9 = declare_stream()
            save(v12, r9)
            r10 = declare_stream()
            save(v13, r10)
            r11 = declare_stream()
            save(v14, r11)
            r12 = declare_stream()
            save(v15, r12)
            r13 = declare_stream()
            save(v16, r13)
            r14 = declare_stream()
            save(v17, r14)
            r15 = declare_stream()
            save(v18, r15)
            r16 = declare_stream()
            save(v19, r16)
            r17 = declare_stream()
            save(v20, r17)
            align()
            assign(v2, (v2+1))
            r1 = declare_stream()
            save(v2, r1)
            align()
            assign(v21, (v21+1))
    with stream_processing():
        r1.buffer(1).save("qm_driver_mock_measurement_shots")
        r2.buffer(100).save("qm_driver_mock_measurement_parity_read_even_ref__p1p2")
        r3.buffer(100).save("qm_driver_mock_measurement_parity_read_even_ref__p7p8")
        r4.buffer(100).save("qm_driver_mock_measurement_parity_read_even_read__p1p2")
        r5.buffer(100).save("qm_driver_mock_measurement_parity_read_even_read__p7p8")
        r6.buffer(100).save("qm_driver_mock_measurement_parity_read_even_diff__p1p2")
        r7.buffer(100).save("qm_driver_mock_measurement_parity_read_even_diff__p7p8")
        r8.buffer(100).save("qm_driver_mock_measurement_parity_read_even_state__p1p2")
        r9.buffer(100).save("qm_driver_mock_measurement_parity_read_even_state__p7p8")
        r10.buffer(100).save("qm_driver_mock_measurement_parity_read_odd_ref__p1p2")
        r11.buffer(100).save("qm_driver_mock_measurement_parity_read_odd_ref__p7p8")
        r12.buffer(100).save("qm_driver_mock_measurement_parity_read_odd_read__p1p2")
        r13.buffer(100).save("qm_driver_mock_measurement_parity_read_odd_read__p7p8")
        r14.buffer(100).save("qm_driver_mock_measurement_parity_read_odd_diff__p1p2")
        r15.buffer(100).save("qm_driver_mock_measurement_parity_read_odd_diff__p7p8")
        r16.buffer(100).save("qm_driver_mock_measurement_parity_read_odd_state__p1p2")
        r17.buffer(100).save("qm_driver_mock_measurement_parity_read_odd_state__p7p8")

config = None

loaded_config = None





from arbok_driver.examples.sub_sequences import CoulombPeaks
from arbok_driver.examples.configurations import (
    coulomb_peaks_set1_set2_conf
)
def test_coulomb_peaks(mock_measurement):
    coulomb_peaks = CoulombPeaks(
        mock_measurement, 'coulomb_peaks', coulomb_peaks_set1_set2_conf
    )
    prg_str = mock_measurement.get_qua_program_as_str(recompile = True)
    assert len(coulomb_peaks.signals) == 3
    assert len(coulomb_peaks.abstract_readouts) == 3
    assert len(coulomb_peaks.readout_groups) == 2



"""Parity read config p1p2 p7p8 for nova."""
from arbok_driver.parameter_types import (
    Time, Voltage, List, Int
)

from arbok_driver.examples.readout_classes import (
    DcAverage,
    Difference,
    Threshold,
) 

from arbok_driver.examples.sub_sequences.parity_readout import ParityRead

parity_read_conf = {
    'sequence': ParityRead,
    'parameters': {
        'gate_elements': {
            'type': List,
            'value': [
                'P1', 'J1',
                'P2',
                'P7', 'J7',
                'P8',
            ]
        },
        'readout_elements': {
            'type': List,
            'value': [
                'SET1', 'SET2'
            ]
        },
        't_wait_home_before': {
            'type': Time,
            'label': 'Wait time at the home point before readout',
            'value': int(100e3/4)
            },
        't_wait_pre_read': {
            'type': Time,
            'value': int(100e3/4)
            },
        't_wait_post_read': {
            'type': Time,
            'value': int(20e3/4)
            },
        't_ramp_to_reference':{
            'type': Time,
            'var_type': 'fixed',
            'value': int(25e3/4)
            },
        't_ramp_to_read': {
            'type': Time,
            'value': int(1e3/4)
            },
        't_wait_after_reset': {
            'type': Time,
            'value': int(4e3/4)
            },
        'v_home': {
            'type': Voltage,
            'elements': {
                    'P1': 0, 'J1': 0,
                    'P2': 0,

                    'P7': 0, 'J7': 0,
                    'P8': 0,
            }
        },
        'v_reference': {
            'type': Voltage,
            'label': 'Default voltage point during the sequence',
            'elements': {
                'P1': 0.023, 'J1': -0.02,
                'P2': -0.023,

                'P7': -0.0325, 'J7': 0.005,
                'P8': 0.0325,
                }
            },
        'v_read': {
            'type': Voltage,
            'label': 'Readout voltage point',
            'elements': {
                'P1': 0.025, 'J1': -0.02,
                'P2': -0.025,

                'P7': -0.035, 'J7': 0.005,
                'P8': 0.035,
            }
        },
    },
    'signals': ['p1p2', 'p7p8'],
    'readout_groups': {
        'ref': {
            'p1p2': {
                'readout_class': DcAverage,
                'signal': 'p1p2',
                'kwargs': {
                    'qua_element': 'SET1'
                },
                'params': {
                    'test': {'type': Int, 'value': 1},
                }
            },
            'p7p8': {
                'readout_class': DcAverage,
                'name': 'ref',
                'signal': 'p7p8',
                'kwargs': {
                    'qua_element': 'SET2'
                }
            }
        },
        'read': {
            'p1p2': {
                'readout_class': DcAverage,
                'signal': 'p1p2',
                'kwargs': {
                    'qua_element': 'SET1'
                }
            },
            'p7p8': {
                'readout_class': DcAverage,
                'signal': 'p7p8',
                'kwargs': {
                    'qua_element': 'SET2'
                }
            }
        },
        'diff': {
            'p1p2': {
                'readout_class': Difference,
                'signal': 'p1p2',
                'kwargs': {
                    'minuend': 'p1p2.ref__p1p2',
                    'subtrahend': 'p1p2.read__p1p2',
                },

            },
            'p7p8': {
                'readout_class': Difference,
                'signal': 'p7p8',
                'kwargs': {
                    'minuend': 'p7p8.ref__p7p8',
                    'subtrahend': 'p7p8.read__p7p8',
                }
            },
        },
        'state': {
            'p1p2': {
                'readout_class': Threshold,
                'signal': 'p1p2',
                'kwargs': {
                    'charge_readout': 'p1p2.diff__p1p2',
                },
                'parameters': {
                    'threshold': {'type': Voltage, 'value': 0.001}
                }
            },
            'p7p8': {
                'readout_class': Threshold,
                'signal': 'p7p8',
                'kwargs': {
                    'charge_readout': 'p7p8.diff__p7p8',
                },
                'parameters': {
                    'threshold': {'type': Voltage, 'value': -0.001}
                }
            },
        },
    },
}


"""Base class for measurement runners."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from abc import ABC, abstractmethod
import logging
import copy
import time
import warnings

import numpy as np
from rich.progress import Progress
import xarray as xr

from arbok_driver.sweep import Sweep

if TYPE_CHECKING:
    from arbok_driver.measurement import Measurement
    from arbok_driver.parameters.sequence_parameter import SequenceParameter

class MeasurementRunnerBase(ABC):
    """
    Helper class constructing QCoDeS measurement loops
    """
    progress_tracker: Progress

    def __init__(
        self,
        measurement: Measurement,
        ext_sweep_list: list[dict[SequenceParameter, np.ndarray]] | None = None,
        register_all: bool = True
        ):
        self.measurement = measurement
        self.arbok_driver = measurement.driver
        self.ext_sweep_list = ext_sweep_list
        self.external_param_values = self._get_external_params(
            self.ext_sweep_list, register_all)

        self.nr_total_batches = self._calculate_nr_total_batches()
        self.inner_func: Callable | None = None
        self.progress_bars = {}
        self.batch_count = 0

        self.opx_dims, self.opx_coords, self.opx_units = \
            generate_dims_and_coords(self.measurement.sweeps)

        self.ext_dims, self.ext_coords, self.ext_units = \
            generate_dims_and_coords(self.ext_sweep_list) if self.ext_sweep_list is not None else ([], {}, {})

        self.dims = self.ext_dims + self.opx_dims
        self.coords = {**self.ext_coords, **self.opx_coords}
        self.units = {**self.ext_units, **self.opx_units}

    @abstractmethod
    def _prepare_measurement(self) -> None:
        """
        Prepare the measurement for running. E.g creating entry in database
        """
        pass

    @abstractmethod
    def _save_results(self, results_xr: xr.Dataset) -> None:
        """
        Save the results of the measurement. E.g. storing data in the database
        """
        pass

    @abstractmethod
    def _handle_keyboard_interrupt(self) -> None:
        """
        Handle keyboard interrupt during the measurement.
        """
        pass

    @abstractmethod
    def _wrap_up_measurement(self) -> None:
        """
        Wrap up the measurement after completion.
        """
        pass

    def run_arbok_measurement(
            self,
            inner_func: Callable | None = None,
            cancle_job_finally: bool = True
            ) -> None:
        """
        Runs the measurement with the given inner function.
        
        Args:
            inner_func (callable): The function to be executed for each measurement
                point. It should accept the datasaver and the current sweep values
                as arguments.
        Returns:
            dataset (Dataset): The QCoDeS dataset containing the measurement results.
        """
        logging.debug("Preparing params and gettables for measurement")
        self.inner_func = inner_func
        self._prepare_measurement()
        try:
            logging.debug("Running measurement with %s", self.measurement.name)
            self.batch_count = 0
            with Progress() as self.progress_tracker:
                self._create_progress_bars()
                self._run_measurement()
        except KeyboardInterrupt:
            logging.debug("Measurement interrupted by user.")
            print("Measurement interrupted by user.")
            self._handle_keyboard_interrupt()
        finally:
            if cancle_job_finally and not self.measurement.driver.is_mock:
                self.measurement.driver.qm_job.cancel()
        self._wrap_up_measurement()

    def _run_measurement(self):
        """
        Builds a native measurement object for the measurement.
        """
        self._create_recursive_measurement_loop(self.ext_sweep_list)
        print("Measurement finished!")

    def _create_recursive_measurement_loop(
            self,
            ext_sweep_list: list[dict[SequenceParameter, np.ndarray]],
            ):
        """
        Creates a recursive measurement loop over the ext_sweep_list.

        Args:
            ext_sweep_list (list[dict]): List of dictionaries containing the sweep
                parameters.
            datasaver (DataSaver): The QCoDeS DataSaver object to save results to.
        """
        # Copy to avoid modifying the original list
        ext_sweep_list = copy.copy(ext_sweep_list)
        if not ext_sweep_list:
            # If the sweep list is empty, execute the inner function
            if self.inner_func is not None:
                self.inner_func()
            ### Program is resumed and all gettables are fetched when ready
            if not self.measurement.driver.is_mock:
                self.measurement.driver.qm_job.resume()
            time.sleep(0.1)
            logging.debug("Job resumed, Fetching gettables")
            self.measurement.wait_until_result_buffer_full(
                (self.progress_bars['batch_progress'], self.progress_tracker)
            )
            batch_coords = {}
            for param, value in self.external_param_values.items():
                dim = self.ext_coords[param.register_name][0]
                batch_coords[param.register_name] = (dim, np.array([value]))
            results_dict = self.measurement.fetch_all_results()
            results_xr = xr.Dataset(coords = batch_coords)
            for gettable, values in results_dict.items():
                dims, coords = gettable.get_full_dims_and_coords(
                    self.ext_dims, batch_coords)
                result_xr = xr.DataArray(
                    data = self._bring_result_to_shape(values),
                    dims = dims,
                    coords = coords
                    )
                results_xr[gettable.register_name] = result_xr
            self.latest_xr_batch = results_xr
            self._save_results(results_xr)
            self.batch_count += 1
            self.update_total_progress_bar()
            return

        ### The first axis will be popped from the list and iterated over
        sweep_dict = ext_sweep_list.pop(0)
        sweep_axis_size = len(list(sweep_dict.values())[0])
        for idx in range(sweep_axis_size):
            ### The parameter values are set for the current iteration
            for param, values in sweep_dict.items():
                value = values[idx]
                logging.debug('Setting %s to %s', param.instrument.name, value)
                param.set(value)
                self.last_updated_dim = param.register_name
                if param in self.external_param_values:
                    self.external_param_values[param] = value
            self._create_recursive_measurement_loop(
                ext_sweep_list=ext_sweep_list)

    def _calculate_nr_total_batches(self) -> int:
        """
        Calculates the total number of batches for the measurement.

        Returns:
            int: The total number of batches.
        """
        if self.ext_sweep_list is not None:
            ext_sweep_lengths = [
                len(next(iter(dic.values()))) for dic in self.ext_sweep_list]
            return int(np.prod(ext_sweep_lengths))
        else:
            self.ext_sweep_list = []
            warnings.warn(
                "NO sweeps outside the OPX registerd. Thus the measurement only"
                " fills the buffer once and finishes "
                "(e.g only lower progres bar)")
            return 1

    def _create_progress_bars(self) -> None:
        """
        Creates progress bars for the measurement.
        """
        total_progress = self.progress_tracker.add_task(
            description = f"[green]Total progress...\n0/{self.nr_total_batches}",
            total = self.nr_total_batches)
        batch_progress = self.progress_tracker.add_task(
            description = "[cyan]Batch progress...",
            total = self.measurement.sweep_size)
        self.progress_bars['total_progress'] = total_progress
        self.progress_bars['batch_progress'] = batch_progress

    def update_total_progress_bar(self) -> None:
        """
        Updates the total progress bar after each batch.
        """
        title = "[green]Total progress\n "
        self.progress_tracker.update(
            self.progress_bars['total_progress'],
            advance=1,
            description=f"{title}{self.batch_count}/{self.nr_total_batches}"
            )
        self.progress_tracker.refresh()

    def _get_external_params(
            self,
            ext_sweep_list: list[dict[SequenceParameter, np.ndarray]] | None,
            register_all: bool
            ) -> dict[SequenceParameter, float | None]:
        """
        Generates a list of external sweep parameters from the ext_sweep_list.
        Args:
            ext_sweep_list (list[dict]): List of dictionaries containing the sweep
                parameters.
        Returns:
            ext_params (list): List of external sweep parameters.
        """
        if ext_sweep_list is None:
            self.ext_sweep_list = []
        else:
            self.ext_sweep_list = ext_sweep_list
        ext_param_values = {}
        for i, sweep_dict in enumerate(self.ext_sweep_list):
            for j, param in enumerate(sweep_dict.keys()):
                if j == 0 or register_all:
                    ext_param_values[param] = None
                else:
                    logging.debug(
                        "Not adding settable %s on axis %s", param.name, i)
        return ext_param_values
    
    def _bring_result_to_shape(self, result: np.ndarray) -> np.ndarray:
        """
        Brings the result DataArray to the correct shape to be added to xarray
        dataset. With each batch we add one datapoint for each external
        dimension setpoint. Therefore wrap the result with singleton dimensions
        for each external dimension.

        Args:
            result (xr.DataArray): The result DataArray to be reshaped.

        Returns:
            xr.DataArray: The reshaped DataArray.
        """
        return np.reshape(result, (1,)*len(self.ext_dims) + result.shape)

def generate_dims_and_coords(
        sweeps: list[dict[SequenceParameter, np.ndarray]] | list[Sweep]
    ) -> tuple[
        list[str],
        dict[str, tuple[str, list]],
        dict[str, str]
        ]:
    """
    Generates a list[str] of dims, coordinates (dict[str, tuple[str, list]])
    and units (dict[str, str]) from a list of sweeps.
    """
    dims = []
    coords = {}
    units = {}
    for sweep in sweeps:
        if isinstance(sweep, Sweep):
            sweep = sweep.config
        for i, (parameter, array) in enumerate(sweep.items()):
            if i == 0:
                dims.append(parameter.register_name)
            coords[parameter.register_name] = (dims[-1], array)
            units[parameter.register_name] = parameter.unit
    return  dims, coords, units



"""Module containing generic StabilityMap class"""
from dataclasses import dataclass
from qm import qua
from arbok_driver import arbok, ReadSequence, ParameterClass
from arbok_driver.parameter_types import Time, ParameterMap, String, Voltage

@dataclass(frozen=True)
class StabilityMapParameters(ParameterClass):
    """Parameter class for StabilityMap read sequence"""
    t_pre_chop: Time
    gate_elements: String
    v_home: ParameterMap[str, Voltage]
    v_level: ParameterMap[str, Voltage]

class StabilityMap(ReadSequence):
    """
    Class containing qua sequence to perform a stability map measurement
    """
    PARAMETER_CLASS = StabilityMapParameters
    arbok_params: StabilityMapParameters
    def __init__(
            self,
            parent,
            name: str,
            sequence_config: dict
    ):
        """
        Constructor method for 'MixedDownUpInit' class

        Args:
            name (str): name of sequence
            device  (Device): Device class for physical device
            sequence_config (dict): config containing pulse parameters
        """
        super().__init__(
            parent = parent,
            name = name,
            sequence_config = sequence_config
        )
        self.elements = list(self.arbok_params.gate_elements.get())
        self.arbok_params.v_home

    def qua_declare(self):
        """
        Setup the python callback
        """
        super().qua_declare()

    def qua_sequence(self):
        qua.align(*self.elements)
        ### Go to point in voltage space
        arbok.ramp(
            elements= self.arbok_params.gate_elements.get(),
            reference = self.arbok_params.v_home,
            target = self.arbok_params.v_level,
            operation = 'unit_ramp',
            )
        qua.wait(self.arbok_params.t_pre_chop.qua, *self.elements)
        qua.align(*self.elements)

        ### Conduct chopped readout(s)
        for _, readout in self.readout_groups["chop"].items():
            readout.qua_measure_and_save()
        qua.align(*self.elements)

        ### Reset elements to home voltage
        arbok.reset_sticky_elements(self.arbok_params.gate_elements.get())




import pytest
from arbok_driver.examples.sub_sequences import (
    SquarePulse, SquarePulseLegacy
)
from arbok_driver.parameter_types import Amplitude, String, Time

def test_sequence_parameter_call(measurement_parameter):
    assert measurement_parameter.qua_var == 'placeholder'
    assert measurement_parameter.qua == 'placeholder'
    assert measurement_parameter() == 'placeholder'
    with pytest.raises(ValueError):
        measurement_parameter(3)

def test_parameter_gettin_in_program(mock_measurement) -> None:
    conf = {
        "parameters": {
            'amplitude': {'type': Amplitude, 'value': 1.5},
            'element': {'type': String, 'value': 'P1'},
            't_ramp': {'type': Time, 'value': int(100)},
            't_square_pulse': {'type': Time, 'value': int(1000)}
        }
    }
    _ = SquarePulse(
        mock_measurement, "square_pulse", conf)
    qua_prog_str = mock_measurement.get_qua_program_as_str()
    qua_prog_str = qua_prog_str.split("program() as prog:")[1]
    mock_measurement.submodules = {}
    mock_measurement._sub_sequences = []
    _ = SquarePulse(
        mock_measurement, "square_pulse", conf)
    qua_prog__legacy_str = mock_measurement.get_qua_program_as_str(
        recompile = True)
    qua_prog__legacy_str = qua_prog__legacy_str.split("program() as prog:")[1]
    assert qua_prog__legacy_str == qua_prog_str



""" Module containing Devices class """
import warnings
from .utils import get_module

class Device():
    """
    Class describing the used device by its config and the used sequence. 
    """
    _master_config: dict
    _param_config: dict
    _default_sequence_configs: dict

    def __init__(
            self, name: str,
            opx_config: dict,
            divider_config: dict,
            master_config: dict | None = None,
            ):
        """
        Constructor class for 'Device' class.

        Args:
            name (str): Name of the used device
            opx_config (dict): Configuration dictionary for the OPX
            divider_config (dict): Configuration dictionary for the divider
            master_config (dict): Configuration dictionary for universally accessible
                parameters across all sub-sequences as well as default
                sequence configurations.
        """
        self.name = name
        self.config = opx_config
        self.elements = list(self.config['elements'].keys())
        self.divider_config = divider_config
        self.master_config = master_config
        self.sub_device_name = None

    @property
    def master_config(self) -> dict:
        return self._master_config
    
    @master_config.setter
    def master_config(self, master_config: dict | None) -> None:
        """
        Update the master config from a dict

        Args:
            master_config (dict): Configuration dictionary for universally accessible
                parameters across all sub-sequences as well as default
                sequence configurations.
        """
        if master_config is None:
            master_config = {}
        if 'parameters' in master_config:
            self._param_config = master_config['parameters']
        else:
            self._param_config = {}
        if 'default_sequence_configs' in master_config:
            self._default_sequence_configs = master_config['default_sequence_configs']
        else:
            self._default_sequence_configs = {}
        self._master_config = master_config

    @property
    def param_config(self):
        return self._param_config
    
    @property
    def default_sequence_configs(self):
        return self._default_sequence_configs



import uuid
from sqlalchemy import (
    BigInteger,
    Boolean,
    ForeignKey,
    Integer,
    JSON,
    String,
    Text,
    UniqueConstraint,
)
from sqlalchemy.dialects.postgresql import ARRAY, JSONB, UUID
from sqlalchemy.orm import (
    DeclarativeBase, Mapped, mapped_column, relationship
)

class Base(DeclarativeBase):
    pass

class SqlRun(Base):
    __tablename__ = "runs"

    ### Run metadata
    run_id: Mapped[int] = mapped_column(
        Integer, primary_key=True, autoincrement=True)
    uuid: Mapped[str] = mapped_column(
        UUID(as_uuid=True), default=uuid.uuid4, unique=True, nullable=False)
    exp_id: Mapped[int] = mapped_column(ForeignKey("experiments.exp_id"))
    experiment = relationship("SqlExperiment", back_populates="runs")
    device_id = mapped_column(ForeignKey("devices.device_id"))
    device = relationship("SqlDevice", back_populates="runs")
    name: Mapped[str] = mapped_column(String)

    setup: Mapped[str] = mapped_column(String)

    ### Data specific to this run
    result_count: Mapped[int] = mapped_column(Integer, default=0)
    batch_count: Mapped[int] = mapped_column(Integer, default=0)
    coords: Mapped[list[str]] = mapped_column(ARRAY(String))
    sweeps: Mapped[dict[int, list[str]]] = mapped_column(JSONB)

    ### Timestamps and status
    start_time: Mapped[int] = mapped_column(BigInteger)
    completed_time: Mapped[int] = mapped_column(
        BigInteger, default=None, nullable=True)
    is_completed: Mapped[bool] = mapped_column(Boolean, default=False)

    measurement_exception: Mapped[str] = mapped_column(
        Text, default=None, nullable=True)
    parent_datasets: Mapped[list[int]] = mapped_column(
        ARRAY(Integer), default=None, nullable=True)


class SqlExperiment(Base):
    __tablename__ = "experiments"

    exp_id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    creation_time: Mapped[int] = mapped_column(BigInteger)
    #run_counter: Mapped[int] = mapped_column(Integer)
    #format_string: Mapped[str] = mapped_column(String)

    runs: Mapped[list[SqlRun]] = relationship(
        "SqlRun", back_populates="experiment", cascade="all, delete-orphan"
    )

    __table_args__ = (
        UniqueConstraint("name", name="uq_experiment_name"),
    )

    def __repr__(self) -> str:
        return (f"<Experiment(exp_id={self.exp_id}, name='{self.name}', "
                f"creation_time={self.creation_time}> ")

class SqlDevice(Base):
    __tablename__ = "devices"

    device_id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    runs: Mapped[list[SqlRun]] = relationship("SqlRun", back_populates="device")

    def __repr__(self) -> str:
        return f"<Device(device_id={self.device_id}, name='{self.name}')>"



"""
Module containing class parity spin initialization of a qubit pair
"""
from dataclasses import dataclass

from qm import qua
from qm.qua._expressions import QuaVariable
from arbok_driver import (
    arbok, ParameterClass, SequenceBase, SubSequence
)
from arbok_driver.parameter_types import (
    Time, ParameterMap, Voltage, List
)

@dataclass(frozen = True)
class ParityInitDeltaParameters(ParameterClass):
    gate_elements: List
    v_home: ParameterMap[str, Voltage]
    v_t1_point: ParameterMap[str, Voltage]
    v_pre_crossing: ParameterMap[str, Voltage]
    v_delta: ParameterMap[str, Voltage]
    t_ramp_to_t1: Time
    t_wait_at_t1: Time
    t_ramp_to_crossing: Time
    t_wait_before_crossing: Time
    t_ramp_over_crossing: Time
    t_wait_after_crossing: Time

class ParityInit(SubSequence):
    """
    Class containing parameters and sequence for parity initialization
    """
    PARAMETER_CLASS= ParityInitDeltaParameters
    arbok_params: ParityInitDeltaParameters

    def __init__(
            self,
            parent: SequenceBase,
            name: str,
            sequence_config: dict
        ):
        """
        Constructor method for 'ParityInit' class
        
        Args:
            parent (InstrumentModule | Instrument): parent sequence
            name (str): name of sequence
            device  (Device): Device class for physical device
            seq_config (dict): config containing pulse parameters
        """
        super().__init__(parent, name, sequence_config)
        self.elements = self.arbok_params.gate_elements.get()

    def qua_sequence(self):
        """QUA sequence to perform odd spin parity initialization (down-up)"""
        ### Ramping to point in the even charge state region to init GS
        arbok.ramp(
			elements= self.elements,
			reference = self.arbok_params.v_home,
			target = self.arbok_params.v_t1_point,
			duration = self.arbok_params.t_ramp_to_t1,
			operation = 'unit_ramp'
			)
        qua.wait(self.arbok_params.t_wait_at_t1.qua, *self.elements)

        ### Ramp to point close to the charge transition
        arbok.ramp(
			elements= self.elements,
			reference = self.arbok_params.v_t1_point,
			target = self.arbok_params.v_pre_crossing,
			duration = self.arbok_params.t_ramp_to_crossing,
			operation = 'unit_ramp'
			)
        qua.wait(self.arbok_params.t_wait_before_crossing.qua, *self.elements)

        ### Ramping (a)diabatically across the charge transition back into the
        ### control region
        arbok.ramp(
			elements= self.elements,
			target = self.arbok_params.v_delta,
			duration = self.arbok_params.t_ramp_over_crossing,
			operation = 'unit_ramp'
			)
        qua.wait(self.arbok_params.t_wait_after_crossing.qua, *self.elements)

        ### Ramping back to the home point resetting elements
        arbok.reset_sticky_elements(self.elements)



"""Module containing reset_sticky_elements function"""
from qm import qua

def reset_sticky_elements(
        element_list: list,
        do_align: bool = True,
        ) -> None:
    """
    Runs `ramp_to_zero` on all sticky elements in the given list.
    This is cruicial to stop numerical errors from accumulation on the sticky
    elements
    """
    for element_name in element_list:
        qua.ramp_to_zero(element_name)
    if do_align:
        qua.align(*element_list)




# Single QUA script generated at 2026-03-28 03:07:59.319027
# QUA library version: 1.2.5


from qm import CompilerOptionArguments
from qm.qua import *

with program() as prog:
    v1 = declare(fixed, )
    v2 = declare(int, )
    a1 = declare(int, value=[10, 20, 30, 40])
    v3 = declare(int, )
    a2 = declare(int, value=[10, 20, 30, 40])
    v4 = declare(int, value=0)
    v5 = declare(int, )
    v6 = declare(int, )
    with infinite_loop_():
        pause()
        assign(v4, 0)
        assign(v5, 0)
        with while_((v5<5)):
            assign(v1, (0.1+Cast.mul_fixed_by_int(0.224775,v5)))
            align()
            assign(v6, 0)
            with while_((v6<4)):
                assign(v2, a1[v6])
                assign(v3, a2[v6])
                align()
                play("ramp"*amp(v1), "gate_1", duration=v3)
                wait(v2, "gate_1")
                play("ramp"*amp(v1), "gate_1", duration=v3)
                align()
                assign(v4, (v4+1))
                r0 = declare_stream()
                save(v4, r0)
                align()
                assign(v6, (v6+1))
            assign(v5, (v5+1))
    with stream_processing():
        r0.buffer(1).save("qm_driver_mock_measurement_shots")

config = None

loaded_config = None




"""Module containing generic sequence for a simple square pulse sequence"""
from dataclasses import dataclass
from qm import qua
from arbok_driver import ParameterClass, SubSequence
from arbok_driver.parameter_types import Amplitude, String, Time

@dataclass(frozen = True)
class SquarePulseParameters(ParameterClass):
    amplitude: Amplitude
    element: String
    t_ramp: Time
    t_square_pulse: Time

class SquarePulse(SubSequence):
    """
    Class containing parameters and sequence for a simple square pulse
    """
    PARAMETER_CLASS = SquarePulseParameters
    arbok_params: SquarePulseParameters
    def qua_sequence(self):
        """Macro that will be played within the qua.program() context"""
        qua.align()
        qua.play(
            pulse = 'ramp'*qua.amp(self.arbok_params.amplitude.qua),
            element = self.arbok_params.element.qua,
            duration = self.arbok_params.t_ramp.qua
            )
        qua.wait(
            self.arbok_params.t_square_pulse.qua,
            self.arbok_params.element.qua)
        qua.play(
            pulse = 'ramp'*qua.amp(self.arbok_params.amplitude.qua),
            element = self.arbok_params.element.qua,
            duration = self.arbok_params.t_ramp.qua
            )
